"""Module containing cython functions, which generate first order 1vN kernel.
   For docstrings see documentation of module neumann1."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import itertools

from ..mytypes import doublenp
from ..mytypes import complexnp

from ..specfuncc cimport func_1vN
from ..aprclass import Approach

cimport numpy as np
cimport cython

ctypedef np.uint8_t bool_t
ctypedef np.int_t int_t
ctypedef np.int64_t long_t
ctypedef np.float64_t double_t
ctypedef np.complex128_t complex_t

@cython.boundscheck(False)
def c_generate_phi1fct(sys):
    cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
    si = sys.si
    cdef np.ndarray[double_t, ndim=1] mulst = sys.leads.mulst
    cdef np.ndarray[double_t, ndim=1] tlst = sys.leads.tlst
    cdef np.ndarray[double_t, ndim=2] dlst = sys.leads.dlst
    #
    cdef long_t c, b, cb
    cdef int_t bcharge, ccharge, charge, l
    cdef double_t Ecb
    #
    cdef int_t nleads = si.nleads
    cdef int_t itype = sys.funcp.itype
    cdef int_t limit = sys.funcp.dqawc_limit
    #
    cdef np.ndarray[complex_t, ndim=3] phi1fct = np.zeros((nleads, si.ndm1, 2), dtype=complexnp)
    cdef np.ndarray[complex_t, ndim=3] phi1fct_energy = np.zeros((nleads, si.ndm1, 2), dtype=complexnp)
    #
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    #
    cdef np.ndarray[complex_t, ndim=1] rez = np.zeros(4, dtype=complexnp)
    #
    for charge in range(si.ncharge-1):
        ccharge = charge+1
        bcharge = charge
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
            Ecb = E[c]-E[b]
            for l in range(nleads):
                func_1vN(Ecb, mulst[l], tlst[l], dlst[l,0], dlst[l,1], itype, limit, rez)
                phi1fct[l, cb, 0] = rez[0]
                phi1fct[l, cb, 1] = rez[1]
                phi1fct_energy[l, cb, 0] = rez[2]
                phi1fct_energy[l, cb, 1] = rez[3]
    sys.phi1fct = phi1fct
    sys.phi1fct_energy = phi1fct_energy
    return 0

#---------------------------------------------------------------------------------------------------
# 1 von Neumann approach
#---------------------------------------------------------------------------------------------------
@cython.boundscheck(False)
def c_generate_kern_1vN(sys):
    cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
    cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
    cdef np.ndarray[complex_t, ndim=3] phi1fct = sys.phi1fct
    si = sys.si
    cdef bint symq = sys.funcp.symq
    cdef long_t norm_rowp = sys.funcp.norm_row
    #
    cdef bool_t bbp_bool, bbpi_bool
    cdef int_t charge, acharge, bcharge, ccharge, l, nleads, \
               aap_sgn, bppbp_sgn, bbpp_sgn, ccp_sgn
    cdef long_t b, bp, bbp, bbpi, bb, \
                a, ap, aap, aapi, \
                bpp, bppbp, bppbpi, bbpp, bbppi, \
                c, cp, ccp, ccpi, \
                bpa, bap, cbp, ba, cb, cpb
    cdef long_t norm_row, last_row, ndm0, npauli,
    cdef complex_t fct_aap, fct_bppbp, fct_bbpp, fct_ccp
    #
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
    cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
    cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
    #
    norm_row = norm_rowp if symq else si.ndm0r
    last_row = si.ndm0r-1 if symq else si.ndm0r
    ndm0, npauli, nleads = si.ndm0, si.npauli, si.nleads
    #
    cdef np.ndarray[double_t, ndim=2] kern = np.zeros((last_row+1, si.ndm0r), dtype=doublenp)
    cdef np.ndarray[double_t, ndim=1] bvec = np.zeros(last_row+1, dtype=doublenp)
    bvec[norm_row] = 1
    for charge in range(si.ncharge):
        acharge = charge-1
        bcharge = charge
        ccharge = charge+1
        for b, bp in itertools.combinations_with_replacement(si.statesdm[bcharge], 2):
            bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
            bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
            if bbp != -1 and bbp_bool:
                bbpi = ndm0 + bbp - npauli
                bbpi_bool = True if bbpi >= ndm0 else False
                if bbpi_bool:
                    kern[bbp, bbpi] = kern[bbp, bbpi] + E[b]-E[bp]
                    kern[bbpi, bbp] = kern[bbpi, bbp] + E[bp]-E[b]
                #--------------------------------------------------
                for a, ap in itertools.product(si.statesdm[acharge], si.statesdm[acharge]):
                    aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
                    if aap != -1:
                        bpa = lenlst[acharge]*dictdm[bp] + dictdm[a] + shiftlst1[acharge]
                        bap = lenlst[acharge]*dictdm[b] + dictdm[ap] + shiftlst1[acharge]
                        fct_aap = 0
                        for l in range(nleads):
                            fct_aap += (+Tba[l, b, a]*Tba[l, ap, bp]*phi1fct[l, bpa, 0].conjugate()
                                        -Tba[l, b, a]*Tba[l, ap, bp]*phi1fct[l, bap, 0])
                        aapi = ndm0 + aap - npauli
                        aap_sgn = +1 if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else -1
                        kern[bbp, aap] = kern[bbp, aap] + fct_aap.imag                              # kern[bbp, aap]   += fct_aap.imag
                        if aapi >= ndm0:
                            kern[bbp, aapi] = kern[bbp, aapi] + fct_aap.real*aap_sgn                # kern[bbp, aapi]  += fct_aap.real*aap_sgn
                            if bbpi_bool:
                                kern[bbpi, aapi] = kern[bbpi, aapi] + fct_aap.imag*aap_sgn          # kern[bbpi, aapi] += fct_aap.imag*aap_sgn
                        if bbpi_bool:
                            kern[bbpi, aap] = kern[bbpi, aap] - fct_aap.real                        # kern[bbpi, aap]  -= fct_aap.real
                #--------------------------------------------------
                for bpp in si.statesdm[bcharge]:
                    bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
                    if bppbp != -1:
                        fct_bppbp = 0
                        for a in si.statesdm[acharge]:
                            bpa = lenlst[acharge]*dictdm[bp] + dictdm[a] + shiftlst1[acharge]
                            for l in range(nleads):
                                fct_bppbp += +Tba[l, b, a]*Tba[l, a, bpp]*phi1fct[l, bpa, 1].conjugate()
                        for c in si.statesdm[ccharge]:
                            cbp = lenlst[bcharge]*dictdm[c] + dictdm[bp] + shiftlst1[bcharge]
                            for l in range(nleads):
                                fct_bppbp += +Tba[l, b, c]*Tba[l, c, bpp]*phi1fct[l, cbp, 0]
                        bppbpi = ndm0 + bppbp - npauli
                        bppbp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else -1
                        kern[bbp, bppbp] = kern[bbp, bppbp] + fct_bppbp.imag                        # kern[bbp, bppbp] += fct_bppbp.imag
                        if bppbpi >= ndm0:
                            kern[bbp, bppbpi] = kern[bbp, bppbpi] + fct_bppbp.real*bppbp_sgn        # kern[bbp, bppbpi] += fct_bppbp.real*bppbp_sgn
                            if bbpi_bool:
                                kern[bbpi, bppbpi] = kern[bbpi, bppbpi] + fct_bppbp.imag*bppbp_sgn  # kern[bbpi, bppbpi] += fct_bppbp.imag*bppbp_sgn
                        if bbpi_bool:
                            kern[bbpi, bppbp] = kern[bbpi, bppbp] - fct_bppbp.real                  # kern[bbpi, bppbp] -= fct_bppbp.real
                    #--------------------------------------------------
                    bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
                    if bbpp != -1:
                        fct_bbpp = 0
                        for a in si.statesdm[acharge]:
                            ba = lenlst[acharge]*dictdm[b] + dictdm[a] + shiftlst1[acharge]
                            for l in range(nleads):
                                fct_bbpp += -Tba[l, bpp, a]*Tba[l, a, bp]*phi1fct[l, ba, 1]
                        for c in si.statesdm[ccharge]:
                            cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
                            for l in range(nleads):
                                fct_bbpp += -Tba[l, bpp, c]*Tba[l, c, bp]*phi1fct[l, cb, 0].conjugate()
                        bbppi = ndm0 + bbpp - npauli
                        bbpp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else -1
                        kern[bbp, bbpp] = kern[bbp, bbpp] + fct_bbpp.imag                           # kern[bbp, bbpp] += fct_bbpp.imag
                        if bbppi >= ndm0:
                            kern[bbp, bbppi] = kern[bbp, bbppi] + fct_bbpp.real*bbpp_sgn            # kern[bbp, bbppi] += fct_bbpp.real*bbpp_sgn
                            if bbpi_bool:
                                kern[bbpi, bbppi] = kern[bbpi, bbppi] + fct_bbpp.imag*bbpp_sgn      # kern[bbpi, bbppi] += fct_bbpp.imag*bbpp_sgn
                        if bbpi_bool:
                            kern[bbpi, bbpp] = kern[bbpi, bbpp] - fct_bbpp.real                     # kern[bbpi, bbpp] -= fct_bbpp.real
                #--------------------------------------------------
                for c, cp in itertools.product(si.statesdm[ccharge], si.statesdm[ccharge]):
                    ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
                    if ccp != -1:
                        cbp = lenlst[bcharge]*dictdm[c] + dictdm[bp] + shiftlst1[bcharge]
                        cpb = lenlst[bcharge]*dictdm[cp] + dictdm[b] + shiftlst1[bcharge]
                        fct_ccp = 0
                        for l in range(nleads):
                            fct_ccp += (+Tba[l, b, c]*Tba[l, cp, bp]*phi1fct[l, cbp, 1]
                                        -Tba[l, b, c]*Tba[l, cp, bp]*phi1fct[l, cpb, 1].conjugate())
                        ccpi = ndm0 + ccp - npauli
                        ccp_sgn = +1 if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else -1
                        kern[bbp, ccp] = kern[bbp, ccp] + fct_ccp.imag                              # kern[bbp, ccp] += fct_ccp.imag
                        if ccpi >= ndm0:
                            kern[bbp, ccpi] = kern[bbp, ccpi] + fct_ccp.real*ccp_sgn                # kern[bbp, ccpi] += fct_ccp.real*ccp_sgn
                            if bbpi_bool:
                                kern[bbpi, ccpi] = kern[bbpi, ccpi] + fct_ccp.imag*ccp_sgn          # kern[bbpi, ccpi] += fct_ccp.imag*ccp_sgn
                        if bbpi_bool:
                            kern[bbpi, ccp] = kern[bbpi, ccp] - fct_ccp.real                        # kern[bbpi, ccp] -= fct_ccp.real
                #--------------------------------------------------
    # Normalisation condition
    kern[norm_row] = np.zeros(si.ndm0r, dtype=doublenp)
    for charge in range(si.ncharge):
        for b in si.statesdm[charge]:
            bb = mapdm0[lenlst[charge]*dictdm[b] + dictdm[b] + shiftlst0[charge]]
            kern[norm_row, bb] += 1
    sys.kern = kern
    sys.bvec = bvec
    return 0

@cython.boundscheck(False)
def c_generate_current_1vN(sys):
    cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
    cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
    cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
    cdef np.ndarray[complex_t, ndim=3] phi1fct = sys.phi1fct
    cdef np.ndarray[complex_t, ndim=3] phi1fct_energy = sys.phi1fct_energy
    si = sys.si
    #
    cdef bool_t bpb_conj, ccp_conj
    cdef int_t bcharge, ccharge, charge, l, nleads,
    cdef long_t c, b, cb, bp, bpb, cp, ccp
    cdef long_t ndm0, ndm1, npauli
    cdef complex_t fct1, fct2, fct1h, fct2h, phi0bpb, phi0ccp
    ndm0, ndm1, npauli, nleads = si.ndm0, si.ndm1, si.npauli, si.nleads
    #
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
    cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
    cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
    #
    cdef np.ndarray[complex_t, ndim=2] phi1 = np.zeros((nleads, ndm1), dtype=complexnp)
    cdef np.ndarray[complex_t, ndim=1] current = np.zeros(nleads, dtype=complexnp)
    cdef np.ndarray[complex_t, ndim=1] energy_current = np.zeros(nleads, dtype=complexnp)
    cdef np.ndarray[complex_t, ndim=1] phi0 = np.zeros(ndm0, dtype=complexnp)
    #
    phi0[0:npauli] = phi0p[0:npauli]
    phi0[npauli:ndm0] = phi0p[npauli:ndm0] + 1j*phi0p[ndm0:]
    #
    for charge in range(si.ncharge-1):
        ccharge = charge+1
        bcharge = charge
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
            for l in range(nleads):
                fct1 = phi1fct[l, cb, 0]
                fct2 = phi1fct[l, cb, 1]
                fct1h = phi1fct_energy[l, cb, 0]
                fct2h = phi1fct_energy[l, cb, 1]
                for bp in si.statesdm[bcharge]:
                    bpb = mapdm0[lenlst[bcharge]*dictdm[bp] + dictdm[b] + shiftlst0[bcharge]]
                    if bpb != -1:
                        bpb_conj = conjdm0[lenlst[bcharge]*dictdm[bp] + dictdm[b] + shiftlst0[bcharge]]
                        phi0bpb = phi0[bpb] if bpb_conj else phi0[bpb].conjugate()
                        phi1[l, cb] = phi1[l, cb] + Tba[l, c, bp]*phi0bpb*fct1
                        current[l] = current[l] + Tba[l, b, c]*Tba[l, c, bp]*phi0bpb*fct1
                        energy_current[l] = energy_current[l] + Tba[l, b, c]*Tba[l, c, bp]*phi0bpb*fct1h
                for cp in si.statesdm[ccharge]:
                    ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
                    if ccp != -1:
                        ccp_conj = conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
                        phi0ccp = phi0[ccp] if ccp_conj else phi0[ccp].conjugate()
                        phi1[l, cb] = phi1[l, cb] + Tba[l, cp, b]*phi0ccp*fct2
                        current[l] = current[l] + Tba[l, b, c]*phi0ccp*Tba[l, cp, b]*fct2
                        energy_current[l] = energy_current[l] + Tba[l, b, c]*phi0ccp*Tba[l, cp, b]*fct2h
    sys.phi1 = phi1
    sys.current = np.array(-2*current.imag, dtype=doublenp)
    sys.energy_current = np.array(-2*energy_current.imag, dtype=doublenp)
    sys.heat_current = sys.energy_current - sys.current*sys.leads.mulst
    return 0

@cython.boundscheck(False)
def c_generate_vec_1vN(np.ndarray[double_t, ndim=1] phi0p, sys):
    #cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
    cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
    cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
    cdef np.ndarray[complex_t, ndim=3] phi1fct = sys.phi1fct
    si = sys.si
    cdef long_t norm_row = sys.funcp.norm_row
    #
    cdef bool_t bbp_bool
    cdef int_t charge, acharge, bcharge, ccharge, l, nleads, \
               aap_sgn, bppbp_sgn, bbpp_sgn, ccp_sgn
    cdef long_t b, bp, bbp, bb, \
                a, ap, aap, \
                bpp, bppbp, bbpp, \
                c, cp, ccp, \
                bpa, bap, cbp, ba, cb, cpb
    cdef long_t ndm0, npauli
    cdef complex_t fct_aap, fct_bppbp, fct_bbpp, fct_ccp, norm
    cdef complex_t phi0aap, phi0bppbp, phi0bbpp, phi0ccp
    ndm0, npauli, nleads = si.ndm0, si.npauli, si.nleads
    #
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
    cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
    cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
    #
    cdef np.ndarray[complex_t, ndim=1] phi0 = np.zeros(ndm0, dtype=complexnp)
    cdef np.ndarray[complex_t, ndim=1] i_dphi0_dt = np.zeros(ndm0, dtype=complexnp)
    #
    phi0[0:npauli] = phi0p[0:npauli]
    phi0[npauli:ndm0] = phi0p[npauli:ndm0] + 1j*phi0p[ndm0:]
    norm = 0
    for charge in range(si.ncharge):
        acharge = charge-1
        bcharge = charge
        ccharge = charge+1
        for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):
            bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
            if bbp != -1:
                if b == bp: norm = norm + phi0[bbp]
                bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
                if bbp_bool:
                    i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + (E[b]-E[bp])*phi0[bbp]
                    #--------------------------------------------------
                    for a, ap in itertools.product(si.statesdm[charge-1], si.statesdm[charge-1]):
                        aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
                        if aap != -1:
                            bpa = lenlst[acharge]*dictdm[bp] + dictdm[a] + shiftlst1[acharge]
                            bap = lenlst[acharge]*dictdm[b] + dictdm[ap] + shiftlst1[acharge]
                            fct_aap = 0
                            for l in range(nleads):
                                fct_aap += (+Tba[l, b, a]*Tba[l, ap, bp]*phi1fct[l, bpa, 0].conjugate()
                                            -Tba[l, b, a]*Tba[l, ap, bp]*phi1fct[l, bap, 0])
                            phi0aap = phi0[aap] if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else phi0[aap].conjugate()
                            i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + fct_aap*phi0aap
                    #--------------------------------------------------
                    for bpp in si.statesdm[charge]:
                        bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
                        if bppbp != -1:
                            fct_bppbp = 0
                            for a in si.statesdm[charge-1]:
                                bpa = lenlst[acharge]*dictdm[bp] + dictdm[a] + shiftlst1[acharge]
                                for l in range(nleads):
                                    fct_bppbp += +Tba[l, b, a]*Tba[l, a, bpp]*phi1fct[l, bpa, 1].conjugate()
                            for c in si.statesdm[charge+1]:
                                cbp = lenlst[bcharge]*dictdm[c] + dictdm[bp] + shiftlst1[bcharge]
                                for l in range(nleads):
                                    fct_bppbp += +Tba[l, b, c]*Tba[l, c, bpp]*phi1fct[l, cbp, 0]
                            phi0bppbp = phi0[bppbp] if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else phi0[bppbp].conjugate()
                            i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + fct_bppbp*phi0bppbp
                        #--------------------------------------------------
                        bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
                        if bbpp != -1:
                            fct_bbpp = 0
                            for a in si.statesdm[charge-1]:
                                ba = lenlst[acharge]*dictdm[b] + dictdm[a] + shiftlst1[acharge]
                                for l in range(nleads):
                                    fct_bbpp += -Tba[l, bpp, a]*Tba[l, a, bp]*phi1fct[l, ba, 1]
                            for c in si.statesdm[charge+1]:
                                cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
                                for l in range(nleads):
                                    fct_bbpp += -Tba[l, bpp, c]*Tba[l, c, bp]*phi1fct[l, cb, 0].conjugate()
                            phi0bbpp = phi0[bbpp] if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else phi0[bbpp].conjugate()
                            i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + fct_bbpp*phi0bbpp
                    #--------------------------------------------------
                    for c, cp in itertools.product(si.statesdm[charge+1], si.statesdm[charge+1]):
                        ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
                        if ccp != -1:
                            cbp = lenlst[bcharge]*dictdm[c] + dictdm[bp] + shiftlst1[bcharge]
                            cpb = lenlst[bcharge]*dictdm[cp] + dictdm[b] + shiftlst1[bcharge]
                            fct_ccp = 0
                            for l in range(nleads):
                                fct_ccp += (+Tba[l, b, c]*Tba[l, cp, bp]*phi1fct[l, cbp, 1]
                                            -Tba[l, b, c]*Tba[l, cp, bp]*phi1fct[l, cpb, 1].conjugate())
                            phi0ccp = phi0[ccp] if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else phi0[ccp].conjugate()
                            i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + fct_ccp*phi0ccp
                    #--------------------------------------------------
    i_dphi0_dt[norm_row] = 1j*(norm-1)
    return np.concatenate((i_dphi0_dt.imag, i_dphi0_dt[npauli:ndm0].real))

class Approach_1vN(Approach):

    kerntype = '1vN'
    generate_fct = c_generate_phi1fct
    generate_kern = c_generate_kern_1vN
    generate_current = c_generate_current_1vN
    generate_vec = c_generate_vec_1vN
#---------------------------------------------------------------------------------------------------



"""Module for constructing many-body quantum dot Hamiltonian and diagonalising it."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from .indexing import empty_szlst
from .indexing import empty_ssqlst
from .indexing import szrange
from .indexing import ssqrange
from .indexing import sz_to_ind
from .indexing import ssq_to_ind

def construct_ham_coulomb(qd, coulomb, statelst, ham_=None):
    """
    Constructs many-body Hamiltonian for given Coulomb matrix elements.

    Parameters
    ----------
    qd : QuantumDot
        QuantumDot object.
    coulomb : dict
        Dictionary containing coulomb matrix elements. The dictionary is of the format
        coulomb[(m, n, k, l)] = U, where m, n, k, l are the state labels
        and U is the strength of the coulomb interaction between these states.
        Note that only the matrix elements k>l, n>m have to be specified.
    statelst : list
        List of indices of states under consideration.
    ham_ : None or array
        nmany by nmany numpy array containing old values of the many-body Hamiltonian.
        The values in coulomb are added to ham_coulomb.

    Returns
    -------
    ham_coulomb : array
        Matrix representation of many-body Coulomb Hamiltonian in Fock basis for states statelst.
    """
    si, mtype = qd.si, qd.mtype
    herm_c, m_less_n = qd.herm_c, qd.m_less_n
    nstates = len(statelst)
    if ham_ is not None:
        ham_coulomb = ham_
    else:
        ham_coulomb = np.zeros((nstates, nstates), dtype=mtype)
    # Iterate over many-body states
    for j1 in range(nstates):
        state = si.get_state(statelst[j1])
        # Iterate over Coulomb matrix elements
        for j2 in coulomb:
            # The suggested convetion is: k>l, n>m
            (m, n, k, l), U = j2, coulomb[j2]
            if state[k] == 1 and state[l] == 1 and k != l:
                # Calculate fermion sign due to two removed electrons in a given state
                # Note that if k is larger than l additional sign appears for flipping k with l
                fsign = np.power(-1, sum(state[0:k])+sum(state[0:l])) * (-1 if k > l else +1)
                statep = list(state)
                statep[k] = 0
                statep[l] = 0
                if statep[m] == 0 and statep[n] == 0 and m != n:
                    # Calculate fermion sign due to two added electrons in a given state
                    # Note that if n is smaller than m additional sign appears for flipping m with n
                    fsign = fsign*np.power(-1, sum(statep[0:n])+sum(statep[0:m])) * (+1 if n > m else -1)
                    statep[m] = 1
                    statep[n] = 1
                    ind = statelst.index(si.get_ind(statep))
                    ham_coulomb[ind, j1] += U*fsign
                    if herm_c and (m != l or n != k) and not (m_less_n and l > k):
                        ham_coulomb[j1, ind] += U.conjugate()*fsign
    return ham_coulomb

def construct_ham_hopping(qd, hsingle, statelst, ham_=None):
    """
    Constructs many-body Hamiltonian for given single-particle hamiltonian.

    Parameters
    ----------
    qd : QuantumDot
        QuantumDot object.
    hsingle : dict
        Dictionary corresponding to single-particle hopping (tunneling) Hamiltonian.
        The dictionary is of the format hsingle[(i, j)] = hij, where i, j are the state labels
        and hij is the matrix element of the single particle Hamiltonian.
    statelst : list
        List of indices of states under consideration.
    ham_ : None or array
        nmany by nmany numpy array containing old values of the many-body Hamiltonian.
        The values in hsingle are added to ham.

    Returns
    -------
    ham_coulomb : array
        Matrix representation of many-body Hamiltonian in Fock basis for states statelst.
    """
    si, mtype, herm_hs = qd.si, qd.mtype, qd.herm_hs
    nstates = len(statelst)
    if ham_ is not None:
        ham_hopping = ham_
    else:
        ham_hopping = np.zeros((nstates, nstates), dtype=mtype)
    # Iterate over many-body states
    for j1 in range(nstates):
        state = si.get_state(statelst[j1])
        # Iterate over single particle Hamiltonian
        for j0 in hsingle:
            (j2, j3), hop = j0, hsingle[j0]
            if j2 == j3:
                if state[j2] == 1:
                    ham_hopping[j1, j1] += hop
            # Remove particle from j2 single particle state, add particle in j3 single particle state
            elif state[j2] == 1 and state[j3] == 0:
                # Calculate fermion sign for added/removed electrons in a given state
                # Note that if j3 is larger than j2 additional sign appears for flipping j3 with j2
                fsign = np.power(-1, sum(state[0:j2])+sum(state[0:j3])) * (+1 if j2 > j3 else -1)
                statep = list(state)
                statep[j2] = 0
                statep[j3] = 1
                ind = statelst.index(si.get_ind(statep))
                ham_hopping[ind, j1] += hop*fsign
                if herm_hs:
                    ham_hopping[j1, ind] += hop.conjugate()*fsign
    return ham_hopping

def construct_manybody_eigenstates(qd, hsingle, coulomb, statelst, ham_=None):
    """
    Calculates eigenstates of many-body Hamiltonian (described by hsingle and coulomb).

    Parameters
    ----------
    qd : QuantumDot
        QuantumDot object.
    hsingle : array
        nsingle by nsingle array corresponding to single-particle hopping (tunneling) Hamiltonian.
    coulomb : list of lists
        List containing coulomb matrix elements.
        coulomb[i] is list of format [m, n, k, l, U], where m, n, k, l are state labels
        and U is strength of coulomb interaction between these states.
        Note that only the matrix elements k>l, n>m have to be specified.
    statelst : list
        List of indices of states under consideration.
    ham_ : None or array
        nmany by nmany numpy array containing old values of the many-body Hamiltonian.

    Returns
    -------
    ham_vals : array
        len(statelst) by 1 array containing eigenvalues of Hamiltonian.
    ham_vecs : array
        len(statelst) by len(statelst) array containing eigenvector matrix of Hamiltonian.
        Columns of ham_vecs correspond to particular eigenvectors.
    """
    if ham_ is not None:
        ham = ham_
    else:
        ham_coulomb = construct_ham_coulomb(qd, coulomb, statelst)
        ham_hopping = construct_ham_hopping(qd, hsingle, statelst)
        ham = ham_coulomb+ham_hopping
    ham_vals, ham_vecs = np.linalg.eigh(ham)
    return ham_vals, ham_vecs

def construct_Ea_manybody(valslst, si):
    """
    Makes a single array corresponding to a given indexing and containing eigenvalues
    of Hamiltonian.

    Parameters
    ----------
    valslst : list of arrays
        List containing eigenvalues. List entry valslst[charge] is an array of definite charge
        eigenvalues.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.

    Returns
    -------
    Ea : array
        nmany by 1 array containing eigenvalues of the Hamiltonian.
    """
    Ea = np.zeros(si.nmany, dtype=float)
    if si.indexing == 'sz':
        # Iterate over charges
        for charge in range(si.ncharge):
            # Iterate over spin projection sz
            for sz in szrange(charge, si.nsingle):
                # Iterate over many-body states for given charge and sz
                szind = sz_to_ind(sz, charge, si.nsingle)
                for ind in range(len(si.szlst[charge][szind])):
                    # The mapping of many-body states is according to szlst
                    Ea[si.szlst[charge][szind][ind]] = valslst[charge][szind][ind]
    elif si.indexing == 'ssq':
        # Iterate over charges
        for charge in range(si.ncharge):
            # Iterate over spin projection sz
            for sz in szrange(charge, si.nsingle):
                szind = sz_to_ind(sz, charge, si.nsingle)
                # Iterate over total spin ssq
                for ssq in ssqrange(charge, sz, si.nsingle):
                    ssqind = ssq_to_ind(ssq, sz)
                    # Iterate over many-body states for given charge, sz, and ssq
                    for ind in range(len(si.ssqlst[charge][szind][ssqind])):
                        # The mapping of many-body states is according to ssqlst
                        Ea[si.ssqlst[charge][szind][ssqind][ind]] = valslst[charge][szind][ssqind][ind]
    else:
        # Iterate over charges
        for charge in range(si.ncharge):
            # Iterate over many-body states for given charge
            for ind in range(len(si.chargelst[charge])):
                # The mapping of many-body states is according to chargelst
                Ea[si.chargelst[charge][ind]] = valslst[charge][ind]
    return Ea

#---------------------------------------------------------------------------------------------------
# manyhamssq.py
def operator_sm(charge, sz, si):
    """
    Construct the operator :math:`S^{-}`, which acts on the states with given :math:`S^{z}`.

    Parameters
    ----------
    charge : int
        States with given charge (number of electrons) are considered.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.

    Returns
    -------
    sm : array
        Matrix (numpy array) describing the specific :math:`S^{-}` operator.
        The operator is expressed in the Fock basis.
    """
    # The minimum possible value of sz for given number of
    # single particle states nsingle and number of particle charge
    szmin = -min(charge, si.nsingle-charge)
    # Acting with S^{-} on the states with smallest possible sz gives 0
    if sz-2 < szmin: return 0
    states = si.get_lst(charge, sz)
    statesm = si.get_lst(charge, sz-2)
    sm = np.zeros((len(statesm), len(states)), dtype=int)
    # States up to index dind have spin up
    # and the states from dind to nsingle have spin down
    dind = int(si.nsingle/2)
    for j1 in range(len(states)):
        state = si.get_state(states[j1])
        for j2 in range(int(si.nsingle/2)):
            if state[j2] == 1 and state[dind+j2] == 0:
                fsign = np.power(-1, sum(state[0:j2])+sum(state[0:dind+j2])+1)
                state2 = list(state)
                # Flip the spin in single particle state from up to down
                state2[j2] = 0
                state2[dind+j2] = 1
                ind = statesm.index(si.get_ind(state2))
                sm[ind, j1] = 2*fsign
    return sm

def operator_sp(charge, sz, si):
    """
    Construct the operator :math:`S^{+}`, which acts on the states with given sz.

    Parameters
    ----------
    charge : int
        States with given charge (number of electrons) are considered.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.

    Returns
    -------
    sp : array
        Matrix (numpy array) describing the specific :math:`S^{+}` operator.
        The operator is expressed in the Fock basis.
    """
    # The maximum possible value of sz for given number of
    # single particle states nsingle and number of particle charge
    szmax = min(charge, si.nsingle-charge)
    # Acting with S^{+} on the states with largest possible sz gives 0
    if sz+2 > szmax: return 0
    states = si.get_lst(charge, sz)
    statesp = si.get_lst(charge, sz+2)
    sp = np.zeros((len(statesp), len(states)), dtype=int)
    # States up to index dind have spin up
    # and the states from dind to nsingle have spin down
    dind = int(si.nsingle/2)
    for j1 in range(len(states)):
        state = si.get_state(states[j1])
        for j2 in range(int(si.nsingle/2)):
            if state[j2] == 0 and state[dind+j2] == 1:
                fsign = np.power(-1, sum(state[0:j2])+sum(state[0:dind+j2]))
                state2 = list(state)
                # Flip the spin in single particle state from down to up
                state2[j2] = 1
                state2[dind+j2] = 0
                ind = statesp.index(si.get_ind(state2))
                sp[ind, j1] = 2*fsign
    return sp

def operator_ssquare(charge, sz, si):
    """
    Construct the operator :math:`S^{2}`, which acts on the states with given sz.
    :math:`S^{2} = (S^{+}S^{-}+S^{-}S^{+})/2+S^{z}S^{z}`

    Parameters
    ----------
    charge : int
        States with given charge (number of electrons) are considered.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.

    Returns
    -------
    ssquare : array
        Matrix (numpy array) describing the specific :math:`S^{+}` operator.
        The operator is expressed in the Fock basis.
    """
    szmin = -min(charge, si.nsingle-charge)
    szmax = min(charge, si.nsingle-charge)
    sp_sm = np.dot(operator_sp(charge, sz-2, si), operator_sm(charge, sz, si)) if sz-2 >= szmin else 0
    sm_sp = np.dot(operator_sm(charge, sz+2, si), operator_sp(charge, sz, si)) if sz+2 <= szmax else 0
    sz_square = np.eye(len(si.get_lst(charge, sz)), dtype=int)*sz**2
    ssquare = (sp_sm+sm_sp)//2+sz_square
    return ssquare

def ssquare_eigenstates(charge, sz, si, prnt=False):
    """
    Find the eigenstates of the operator :math:`S^{2}` for given sz.

    Parameters
    ----------
    charge : int
        States with given charge (number of electrons) are considered.
    sz : int
        Value of sz.
        sz odd corresponds to fractional spin. For example, 1->1/2, 3->3/2, etc.
        sz even corresponds to integer spin. For example, 0->0, 2->1, etc.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.
    prnt : bool
        If true, then print the eigenstates.

    Returns
    -------
    ssq_eigvec : list of arrays
        List of numpy arrays containing the eingenstates of :math:`S^{2}`.
        ssq_eigvec[ssqind] gives the eigenstate matrix for :math:`S^{2}` =ssq.
    """
    if charge%2 != sz%2:
        print("WARNING: charge and sz need to have the same parity. Return 0.")
        return 0
    ssq_eigvec = [0 for j in ssqrange(charge, sz, si.nsingle)]
    ssquare = operator_ssquare(charge, sz, si)
    eigval, eigvec = np.linalg.eigh(ssquare)
    eigval = np.rint(-1 + np.sqrt(1+eigval))
    for ssq in ssqrange(charge, sz, si.nsingle):
        v = np.nonzero(eigval == ssq)
        ssqind = ssq_to_ind(ssq, sz)
        ssq_eigvec[ssqind] = eigvec[:][:, v[0][0]:v[0][-1]+1]
        if prnt:
            print('-'*30)
            print('charge =', charge, ', sz =', sz, ', ssq =', ssq)
            print(ssq_eigvec[ssqind])
    return ssq_eigvec

def ssquare_all_szlow(si, prnt=False):
    """
    Find the eigenstates of the operator :math:`S^{2}` for sz=0 or 1.

    Parameters
    ----------
    si : StateIndexing
        StateIndexing or StateIndexingDM object.
    prnt : bool
        If true, then print the eigenstates.

    Returns
    -------
    ssq_eigvec : list of arrays
        List of numpy arrays containing the eingenstates of :math:`S^{2}` for lowest abs(sz) value
        possible.
        ssq_eigvec[charge][ssq] gives the eigenstate matrix for given charge and :math:`S^{2}` =ssq.
    """
    ssq_eigvec_all_szlow = [0 for i in range(si.ncharge)]
    for charge in range(si.ncharge):
        ssq_eigvec_all_szlow[charge] = ssquare_eigenstates(charge, -(charge%2), si)
    if prnt:
        for charge in range(si.ncharge):
            for ssq in ssqrange(charge, -(charge%2), si.nsingle):
                print('-'*30)
                print('charge =', charge, ', sz =', -(charge%2), ', ssq =', ssq)
                print(ssq_eigvec_all_szlow[charge][ssq//2])
    return ssq_eigvec_all_szlow

def construct_manybody_eigenstates_ssq(qd, charge, sz, ssq, hsingle, coulomb, ham_=None):
    """
    Find the many-body eigenstates of the Hamiltonian, which are also the eigenstates of sz
    and :math:`S^{2}` for given charge, sz and :math:`S^{2}`.

    Parameters
    ----------
    qd : QuantumDot
        QuantumDot object.
    charge : int
        Charge (number of particles) of the states to consider.
    sz : int
        sz value of the states to consider.
    ssq : int
        :math:`S^{2}` value of the states to consider.
    hsingle : dict
        Dictionary corresponding to single-particle hopping (tunneling) Hamiltonian.
        The dictionary is of the format hsingle[(i, j)] = hij, where i, j are the state labels
        and hij is the matrix element of the single particle Hamiltonian.
    coulomb : dict
        Dictionary containing coulomb matrix elements. The dictionary is of the format
        coulomb[(m, n, k, l)] = U, where m, n, k, l are the state labels
        and U is the strength of the coulomb interaction between these states.
        Note that only the matrix elements k>l, n>m have to be specified.
    ham_ : None or array
        nmany by nmany numpy array containing old values of the many-body Hamiltonian.
        The values in hsingle and coulomb are added to ham.

    Returns
    -------
    eigvalp : array
        Eigenvalues of the Hamltonian for given charge, sz, :math:`S^{2}`.
    eigvecssq : array
        Corresponding eigenvectors of the Hamiltonian.
    """
    si = qd.si
    statelst = si.get_lst(charge=charge, sz=sz)
    if ham_ is not None:
        ham = ham_
    else:
        ham = (construct_ham_coulomb(qd, coulomb, statelst)
              +construct_ham_hopping(qd, hsingle, statelst))
    ssq_eigvec = ssquare_eigenstates(charge, sz, si)
    ssqind = ssq_to_ind(ssq, sz)
    # Write the Hamiltonian in the S^{2} basis for given value of S^{2}=ssq
    hamp = np.dot(ssq_eigvec[ssqind].T, np.dot(ham,  ssq_eigvec[ssqind]))
    eigvalp, eigvecp = np.linalg.eigh(hamp)
    eigvecssq = np.dot(ssq_eigvec[ssqind], eigvecp)
    return eigvalp, eigvecssq

def construct_manybody_eigenstates_ssq_all(qd, charge, hsingle, coulomb, ham_=None):
    """
    Find the many-body eigenstates of the Hamiltonian for given charge.
    The eigenstates are also the eigenstates of sz and :math:`S^{2}`.

    Parameters
    ----------
    qd : QuantumDot
        QuantumDot object.
    charge : int
        Charge (number of particles) of the states to consider.
    hsingle : dict
        Dictionary corresponding to single-particle hopping (tunneling) Hamiltonian.
        The dictionary is of the format hsingle[(i, j)] = hij, where i, j are the state labels
        and hij is the matrix element of the single particle Hamiltonian.
    coulomb : dict
        Dictionary containing coulomb matrix elements. The dictionary is of the format
        coulomb[(m, n, k, l)] = U, where m, n, k, l are the state labels
        and U is the strength of the coulomb interaction between these states.
        Note that only the matrix elements k>l, n>m have to be specified.
    ham_ : None or array
        nmany by nmany numpy array containing old values of the many-body Hamiltonian.
        The values in hsingle and coulomb are added to ham.

    Returns
    -------
    valslst : list of arrays
        Eigenvalues of the Hamltonian for given charge.
        valslst[szind][ssqind] are the eigenvalues for given sz and :math:`S^{2}`.
    vecslst : list of arrays
        Eigenvectors corresponding to valslst.
    """
    si = qd.si
    szlow = -(charge%2)
    nsingle = si.nsingle
    #valslst = [[None] for k in ssqrange(charge, szlow, nsingle)]
    valslst = [[[None] for k in ssqrange(charge, j, nsingle)] for j in szrange(charge, nsingle)]
    vecslst = [[[None] for k in ssqrange(charge, j, nsingle)] for j in szrange(charge, nsingle)]
    #
    statelst = si.get_lst(charge=charge, sz=szlow)
    if ham_ is not None:
        ham = ham_
    else:
        ham = (construct_ham_coulomb(qd, coulomb, statelst)
              +construct_ham_hopping(qd, hsingle, statelst))
    #NOTE: Can be an argument of the function instead
    ssq_eigvec = ssquare_eigenstates(charge, szlow, si)
    #
    # Find eigenstates corresponding to sz=1 or 0
    szind = sz_to_ind(szlow, charge, nsingle)
    for ssq in ssqrange(charge, szlow, nsingle):
        ssqind = ssq_to_ind(ssq, szlow)
        hamp = np.dot(ssq_eigvec[ssqind].T, np.dot(ham,  ssq_eigvec[ssqind]))
        eigvalp, eigvecp = np.linalg.eigh(hamp)
        eigvecssq = np.dot(ssq_eigvec[ssqind], eigvecp)
        valslst[szind][ssqind] = eigvalp
        vecslst[szind][ssqind] = eigvecssq
    # Construct the eigenstates with larger (positive) sz by acting with S^{+}
    szmax = min(charge, si.nsingle-charge)
    for sz in range(szlow, szmax-1, 2):
        szind = sz_to_ind(sz, charge, nsingle)
        szind2 = sz_to_ind(sz+2, charge, nsingle)
        for ssq in ssqrange(charge, sz+2, nsingle):
            ssqind = ssq_to_ind(ssq, sz)
            ssqind2 = ssq_to_ind(ssq, sz+2)
            sp = operator_sp(charge, sz, si)
            fct = np.sqrt(ssq*(ssq+2) - sz*(sz+2))
            vecslst[szind2][ssqind2] = np.dot(sp, vecslst[szind][ssqind])/fct
            valslst[szind2][ssqind2] = valslst[szind][ssqind]
    # Construct the eigenstates with smaller (negative) sz by acting with S^{-}
    for sz in range(szlow, -szmax+1, -2):
        szind = sz_to_ind(sz, charge, nsingle)
        szind2 = sz_to_ind(sz-2, charge, nsingle)
        for ssq in ssqrange(charge, sz-2, nsingle):
            ssqind = ssq_to_ind(ssq, sz)
            ssqind2 = ssq_to_ind(ssq, sz-2)
            sm = operator_sm(charge, sz, si)
            fct = np.sqrt(ssq*(ssq+2) - sz*(sz-2))
            vecslst[szind2][ssqind2] = np.dot(sm, vecslst[szind][ssqind])/fct
            valslst[szind2][ssqind2] = valslst[szind][ssqind]
    return valslst, vecslst
#---------------------------------------------------------------------------------------------------

def make_hsingle_mtr(hsingle, nsingle, mtype=float):
    """
    Makes single particle Hamiltonian matrix from a list or a dictionary.

    Parameters
    ----------
    hsingle : list or dictionary
        Contains single particle tunneling amplitudes
    nsingle : int
        Number of single particle states.
    mtype : type
        Defines type of tleads matrix. For example, float, complex, etc.

    Returns
    -------
    hsingle_mtr : array
        hsingle by hsingle numpy array containing single particle Hamiltonian.
    """
    hsingle_mtr = np.zeros((nsingle, nsingle), dtype=mtype)
    for j0 in hsingle:
        if isinstance(hsingle, list):    j1, j2, hop = j0
        elif isinstance(hsingle, dict): (j1, j2), hop = j0, hsingle[j0]
        hsingle_mtr[j1, j2] += hop
        if j1 != j2:
            hsingle_mtr[j2, j1] += hop.conjugate()
    return hsingle_mtr

def make_hsingle_dict(qd, hsingle):
    """
    Makes single particle Hamiltonian dictionary.

    Parameters
    ----------
    hsingle : list, dict, or array
        Contains single particle Hamiltonian.
    qd : QuantumDot
        QuantumDot object.

    Returns
    -------
    hsingle_dict : dictionary
        Dictionary containing single particle Hamiltonian.
        hsingle[(state1, state2)] gives the matrix element.
    """
    si = qd.si
    #
    if isinstance(hsingle, dict):
        hsingle_dict = hsingle
    elif isinstance(hsingle, list):
        hsingle_dict = {}
        for j0 in hsingle:
            j1, j2, hop = j0
            hsingle_dict.update({(j1, j2):hop})
    elif isinstance(hsingle, np.ndarray):
        nsingle = hsingle.shape[0]
        hsingle_dict = {}
        herm_hs = qd.herm_hs
        for j1 in range(nsingle):
            for j2 in range(j1 if herm_hs else 0, nsingle):
                if hsingle[j1, j2] != 0:
                    hsingle_dict.update({(j1, j2):hsingle[j1, j2]})
    #
    if si.symmetry is 'spin':
        hsingle_dict_spin = dict(hsingle_dict)
        for j0 in hsingle_dict:
            j1, j2 = j0
            hop = hsingle_dict[j0]
            hsingle_dict_spin.update({(j1+si.nsingle_sym,
                                       j2+si.nsingle_sym):hop})
        return hsingle_dict_spin
    else:
        return hsingle_dict

def make_coulomb_dict(qd, coulomb):
    """
    Makes Coulomb matrix element dictionary.

    Parameters
    ----------
    coulomb : list, dict, or array
        Contains coulomb matrix elements.
    qd : QuantumDot
        QuantumDot object.

    Returns
    -------
    coulomb_dict : dictionary
        Dictionary containing coulomb matrix element.
        coulomb[(state1, state2, state3, state4)] gives the coulomb matrix element U.
    """
    si = qd.si
    #
    if isinstance(coulomb, dict):
        coulomb_dict = coulomb
    elif isinstance(coulomb, (list, np.ndarray)):
        coulomb_dict = {}
        for j0 in coulomb:
            m, n, k, l, U = j0
            if U != 0:
                coulomb_dict.update({(m, n, k, l):U})
    #
    if si.symmetry is 'spin':
        coulomb_dict_spin = dict(coulomb_dict)
        nss = si.nsingle_sym
        herm_c, m_less_n = qd.herm_c, qd.m_less_n
        for j0 in coulomb_dict:
            m, n, k, l = j0
            U = coulomb_dict[j0]
            # up, down, down, up
            if not herm_c or (l, k+nss, n+nss, m) not in coulomb_dict_spin.keys():
                coulomb_dict_spin.update({(m, n+nss, k+nss, l):U})
            # down, up, up, down
            if m_less_n:
                if not herm_c or (k, l+nss, m+nss, n) not in coulomb_dict_spin.keys():
                    coulomb_dict_spin.update({(n, m+nss, l+nss, k):U})
            else:
                coulomb_dict_spin.update({(m+nss, n, k, l+nss):U})
            # down, down, down, down
            coulomb_dict_spin.update({(m+nss, n+nss, k+nss, l+nss):U})
        return coulomb_dict_spin
    else:
        return coulomb_dict
#---------------------------------------------------------------------------------------------------

class QuantumDot(object):
    """
    Class for constructing and diagonalising many-body Hamiltonian describing
    the quantum dot.

    Attributes
    ----------
    hsingle : dict, list, or array
        Dictionary, list, or array corresponding to single-particle hopping (tunneling) Hamiltonian.
        On input list or array gets converted to dictionary.
    coulomb : dict, list, or array
        Dictionary, list, or array or containing coulomb matrix elements.
        For dictionary:    coulomb[(m, n, k, l)] = U, where m, n, k, l are the state labels.
        For list or array: coulomb[i] is list of the format [m, n, k, l, U].
        U is the strength of the coulomb interaction between the states (m, n, k, l).
        Note that only the matrix elements k>l, n>m have to be specified.
        On input list or array gets converted to dictionary.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.
    herm_hs : bool
        When herm_hs=True the conjugated elements are added automatically to many-body Hamiltonian
        from hsingle. For example, when hsingle={(0,1): val} the element {(1,0): val.conjugate()}
        will also be added.
    herm_c : bool
        When herm_c=True the conjugated elements are added automatically to many-body Hamiltonian
        from coulomb. For example, when coulomb={(0,2,3,1): val} the element
        {(1,3,2,0): val.conjugate()} will also be added. However, we note that for m_less_n=True
        if coulomb={(0,2,1,3): val} the element (3,1,2,0) will not be included.
    m_less_n : bool
        When m_less_n=True the coulomb matrix element (m, n, k, l) has to have m<n.
    mtype : type
        Defines the type of the Hamiltonian matrices. For example, float, complex, etc.
    ham_ssq_q : bool
        If ham_ssq_q is False then for 'ssq' indexing only Hamiltonian with lowest abs(sz) value
        is stored.
    valslst : list of arrays
        List containing ncharge entries, where valslst[charge] is an array having Hamiltonian
        eigenvalues for definite charge.
    vecslst : list of arrays
        List containing ncharge entries, where vecslst[charge] is an array corresponding to
        Hamiltonian eigenvectors for definite charge.
    hamlst : list of arrays
        List containing ncharge entries, where hamlst[charge] is an array corresponding to many-body
        Hamiltonian for definite charge.
    Ea : array
        nmany by 1 array containing many-body Hamiltonian eigenvalues.
    """

    def __init__(self, hsingle, coulomb, si,
                 herm_hs=True, herm_c=False, m_less_n=True,
                 mtype=float, ham_ssq_q=False):
        """Initialization of the QuantumDot class."""
        self.si = si
        self.mtype = mtype
        self.ham_ssq_q = ham_ssq_q
        self.herm_hs = herm_hs
        self.herm_c = herm_c
        self.m_less_n = m_less_n
        self.hsingle = make_hsingle_dict(self, hsingle)
        self.coulomb = make_coulomb_dict(self, coulomb)
        if si.indexing == 'sz':
            self.valslst = empty_szlst(si.nsingle, True)
            self.vecslst = empty_szlst(si.nsingle, True)
            self.hamlst = empty_szlst(si.nsingle, True)
        elif si.indexing == 'ssq':
            self.valslst = empty_ssqlst(si.nsingle, True)
            self.vecslst = empty_ssqlst(si.nsingle, True)
            self.hamlst = empty_szlst(si.nsingle, True)
        else:
            self.valslst = [None]*si.ncharge
            self.vecslst = [None]*si.ncharge
            self.hamlst = [None]*si.ncharge
        self.add(self.hsingle, self.coulomb, False)
        self.Ea = np.zeros(si.nmany, dtype=float)
        self.Ea_ext = None

    def add(self, hsingle=None, coulomb=None, updateq=True):
        """
        Adds a value to single particle Hamiltonian and Coulomb matrix elements
        and correspondingly redefines many-body Hamiltonian.

        Parameters
        ----------
        hsingle : dict
            Dictionary describing what single-particle Hamiltonian values to add.
            For example, hsingle[(state1, state2)] = value to add.
        coulomb : dict
            Dictionary describing what Coulomb matrix element values to add.
            For example, coulomb[(state1, state2, state3, state4)] = value to add.
        updateq : bool
            Specifies if the values of the single particle Hamiltonian
            and the Coulomb matrix elements will be updated.
            The many-body Hamiltonian will be updates in either case.
        """
        if updateq:
            hsingle = {} if hsingle is None else make_hsingle_dict(self, hsingle)
            coulomb = {} if coulomb is None else make_coulomb_dict(self, coulomb)
            for j0 in hsingle:
                try:    self.hsingle[j0] += hsingle[j0]       # if hsingle[j0] != 0:
                except: self.hsingle.update({j0:hsingle[j0]}) # if hsingle[j0] != 0:
            for j0 in coulomb:
                try:    self.coulomb[j0] += coulomb[j0]       # if coulomb_diff != 0:
                except: self.hsingle.update({j0:coulomb[j0]}) # if coulomb_diff != 0:
        #
        if self.si.indexing == 'sz':
            for charge in range(self.si.ncharge):
                for sz in szrange(charge, self.si.nsingle):
                    szlst = self.si.get_lst(charge, sz)
                    szind = sz_to_ind(sz, charge, self.si.nsingle)
                    self.hamlst[charge][szind] = construct_ham_coulomb(self, coulomb, szlst,
                                                                       self.hamlst[charge][szind])
                    self.hamlst[charge][szind] = construct_ham_hopping(self, hsingle, szlst,
                                                                       self.hamlst[charge][szind])
        elif self.si.indexing == 'ssq':
            for charge in range(self.si.ncharge):
                szlow = -(charge%2)
                szrng = szrange(charge, self.si.nsingle) if self.ham_ssq_q else [szlow]
                for sz in szrng:
                    szlst = self.si.get_lst(charge, sz)
                    szind = sz_to_ind(sz, charge, self.si.nsingle)
                    self.hamlst[charge][szind] = construct_ham_coulomb(self, coulomb, szlst,
                                                                       self.hamlst[charge][szind])
                    self.hamlst[charge][szind] = construct_ham_hopping(self, hsingle, szlst,
                                                                       self.hamlst[charge][szind])
        else:
            for charge in range(self.si.ncharge):
                chargelst = self.si.get_lst(charge)
                self.hamlst[charge] = construct_ham_coulomb(self, coulomb, chargelst,
                                                            self.hamlst[charge])
                self.hamlst[charge] = construct_ham_hopping(self, hsingle, chargelst,
                                                            self.hamlst[charge])

    def change(self, hsingle=None, coulomb=None, updateq=True):
        """
        Changes the values of the single particle Hamiltonian and Coulomb matrix elements
        and correspondingly redefines many-body Hamiltonian.

        Parameters
        ----------
        hsingle : dict
            Dictionary describing what single-particle Hamiltonian values to add.
            For example, hsingle[(state1, state2)] = value to add.
        coulomb : dict
            Dictionary describing what Coulomb matrix element values to add.
            For example, coulomb[(state1, state2, state3, state4)] = value to add.
        updateq : bool
            Specifies if the values of the single particle Hamiltonian
            and the Coulomb matrix elements will be updated.
            The many-body Hamiltonian will be updates in either case.
        """
        hsingle = {} if hsingle is None else make_hsingle_dict(self, hsingle)
        coulomb = {} if coulomb is None else make_coulomb_dict(self, coulomb)
        #
        hsingle_add = {}
        coulomb_add = {}
        # Find the differences from the previous single-particle Hamiltonian
        # and from the previous Coulomb matrix elements
        for j0 in hsingle:
            try:
                hsingle_diff = hsingle[j0]-self.hsingle[j0]
                if hsingle_diff != 0:
                    hsingle_add.update({j0:hsingle_diff})
                    if updateq: self.hsingle[j0] += hsingle_diff
            except:
                hsingle_diff = hsingle[j0]
                if hsingle_diff != 0:
                    hsingle_add.update({j0:hsingle_diff})
                    if updateq: self.hsingle.update({j0:hsingle_diff})
        for j0 in coulomb:
            try:
                coulomb_diff = coulomb[j0]-self.coulomb[j0]
                if coulomb_diff != 0:
                    coulomb_add.update({j0:coulomb_diff})
                    if updateq: self.coulomb[j0] += coulomb_diff
            except:
                coulomb_diff = coulomb[j0]
                if coulomb_diff != 0:
                    coulomb_add.update({j0:coulomb_diff})
                    if updateq: self.coulomb.update({j0:coulomb_diff})
        # Then add the differences
        self.add(hsingle_add, coulomb_add, False)

    def diagonalise_charge(self, charge='n', sz='n', ssq='n'):
        """Diagonalises Hamiltonian for given charge.
           Additionally, spin projection 'sz' and toltal spin 'ssq' can be specified
           if those symmetries exist for considered model.

        Parameters
        ----------
        charge : int
            Charge under consideration.
        sz : int
            Spin projection.
        ssq : int
            Total spin.
        """
        if charge == 'n':
            return None
        elif sz == 'n':
            chargelst = self.si.get_lst(charge)
            (self.valslst[charge],
             self.vecslst[charge]) = (
                construct_manybody_eigenstates(self, self.hsingle, self.coulomb,
                                               chargelst, self.hamlst[charge]) )
            return self.valslst[charge], self.vecslst[charge]
        elif ssq == 'n' and (self.si.indexing == 'sz' or self.si.indexing == 'ssq'):
            szlst = self.si.get_lst(charge, sz)
            szind = sz_to_ind(sz, charge, self.si.nsingle)
            (self.valslst[charge][szind],
             self.vecslst[charge][szind]) = (
                construct_manybody_eigenstates(self, self.hsingle, self.coulomb,
                                               szlst, self.hamlst[charge][szind]) )
            return self.valslst[charge][szind], self.vecslst[charge][szind]
        elif self.si.indexing == 'ssq':
            szind = sz_to_ind(sz, charge, self.si.nsingle)
            ssqind = ssq_to_ind(ssq, sz)
            (self.valslst[charge][szind][ssqind],
             self.vecslst[charge][szind][ssqind]) = (
                construct_manybody_eigenstates_ssq(self, charge, sz, ssq,
                                                   self.hsingle, self.coulomb,
                                                   self.hamlst[charge][szind]) )
            return self.valslst[charge][szind][ssqind], self.vecslst[charge][szind][ssqind]
        else:
            print("WARNING: No indexing by 'sz' or 'ssq'. Return None.")
            return None

    def diagonalise(self):
        """Diagonalises Hamiltonians for all charge states."""
        if self.si.indexing == 'sz':
            for charge in range(self.si.ncharge):
                for sz in szrange(charge, self.si.nsingle):
                    self.diagonalise_charge(charge, sz)
        elif self.si.indexing == 'ssq':
            for charge in range(self.si.ncharge):
                szlow = -(charge%2)
                szind = sz_to_ind(szlow, charge, self.si.nsingle)
                (self.valslst[charge], self.vecslst[charge]) = (
                    construct_manybody_eigenstates_ssq_all(self, charge, self.hsingle, self.coulomb,
                                                           self.hamlst[charge][szind]) )
        else:
            for charge in range(self.si.ncharge):
                self.diagonalise_charge(charge)
        self.set_Ea()

    def set_Ea(self):
        """Sets the many-body eigenstates using construct_Ea_manybody()."""
        self.Ea = construct_Ea_manybody(self.valslst, self.si)



# -- Paths ------------------------------------------------
import sys
import os

sys.path.insert(0, os.path.abspath('../../qmeq/'))

# -- General configuration ------------------------------------------------
needs_sphinx = '1.3'

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.mathjax',
              'sphinx.ext.napoleon'
             ]

templates_path = ['_templates']
#source_suffix = ['.rst', '.md']
source_suffix = '.rst'
#source_encoding = 'utf-8-sig'
master_doc = 'index'

project = u'qmeq'
copyright = u'2016, Gediminas Kirsanskas'
author = u'Gediminas Kirsanskas'

version = u'1.0'
release = u'1.0'

language = None
#today = ''
#today_fmt = '%B %d, %Y'
exclude_patterns = []
#default_role = None
#add_function_parentheses = True
#add_module_names = True
#show_authors = False
pygments_style = 'sphinx'
#modindex_common_prefix = []
#keep_warnings = False
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------
#html_theme = 'alabaster'
#html_theme = 'classic'
html_theme = 'sphinx_rtd_theme'
#html_theme = 'sphinxdoc'
#html_theme = 'epub'
#html_theme_options = {}
#html_theme_path = []
#html_title = u'qmeq v0'
#html_short_title = None
#html_logo = None
#html_favicon = None
#html_static_path = ['_static']
#html_extra_path = []
#html_last_updated_fmt = None
#html_use_smartypants = True
#html_sidebars = {}
#html_additional_pages = {}
#html_domain_indices = True
#html_use_index = True
#html_split_index = False
#html_show_sourcelink = True
#html_show_sphinx = True
#html_show_copyright = True
#html_use_opensearch = ''
#html_file_suffix = None
#html_search_language = 'en'
#html_search_options = {'type': 'default'}
#html_search_scorer = 'scorer.js'
htmlhelp_basename = 'qmeqdoc'


# -- Options for LaTeX output ---------------------------------------------
latex_elements = { }
latex_documents = [
    (master_doc, 'qmeq.tex', u'qmeq Documentation',
     u'Gediminas Kirsanskas', 'manual'),
]
#latex_logo = None
#latex_use_parts = False
#latex_show_pagerefs = False
#latex_show_urls = False
#latex_appendices = []
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------
man_pages = [
    (master_doc, 'qmeq', u'qmeq Documentation',
     [author], 1)
]
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------
texinfo_documents = [
    (master_doc, 'qmeq', u'qmeq Documentation',
     author, 'qmeq', 'One line description of project.',
     'Miscellaneous'),
]
#texinfo_appendices = []
#texinfo_domain_indices = True
#texinfo_show_urls = 'footnote'
#texinfo_no_detailmenu = False



import numpy as np
from numpy.linalg import norm
from qmeq.builder import *
import qmeq
import itertools

EPS = 1e-11
CHECK_PY = False
PRNTQ = False
SERIAL_TRIPLE_DOT = False

class Parameters_double_dot_spinful(object):

    def __init__(self):
        tL, tR = 2.0, 1.0
        self.tleads = {(0,0): tL, (2,2): tL,
                       (1,1): tR, (3,3): tR,
                       (0,1): 0.3*tL, (2,3): 0.3*tL,
                       (1,0): 0.1*tR, (3,2): 0.1*tR}
        #
        e1, e2, omega = -10, -12, 20
        self.hsingle = {(0,0):e1, (2,2):e1,
                        (1,1):e2, (3,3):e2,
                        (0,1):omega, (2,3):omega}
        uintra, uinter = 80, 30
        self.coulomb = {(0,2,2,0):uintra, (1,3,3,1):uintra,
                        (0,1,1,0):uinter, (0,3,3,0):uinter,
                        (1,2,2,1):uinter, (2,3,3,2):uinter}
        #
        vbias, temp, dband = 5.0, 50.0, 1000.0
        self.mulst = [vbias/2, -vbias/2, vbias/2, -vbias/2]
        self.tlst = [temp, temp, temp, temp]
        self.dlst = [dband, dband, dband, dband]
        self.nsingle, self.nleads = 4, 4

class Parameters_double_dot_spinless(object):

    def __init__(self):
        tL, tR = 2.0, 1.0
        self.tleads = {(0,0): tL, (1,1): tR}
        #
        e1, e2, omega = -10, -12, 20
        self.hsingle = {(0,0):e1, (1,1):e2, (0,1):omega}
        uinter = 30.
        self.coulomb = {(0,1,1,0):uinter}
        #
        vbias, temp, dband = 5.0, 25.0, 1000.0
        self.mulst = [vbias/2, -vbias/2]
        self.tlst = [temp, temp]
        self.dlst = [dband, dband]
        #
        self.nsingle, self.nleads = 2, 2
        self.kpnt = np.power(2, 9)

class Parameters_single_orbital_spinful(object):

    def __init__(self):
        tL, tR = 2.0, 1.0
        self.tleads = {(0, 0):tL, (1, 0):tR,
                       (2, 1):tL, (3, 1):tR}
        #
        vgate = 0.0
        self.hsingle = {(0, 0): vgate,
                        (1, 1): vgate}
        #
        U = 300.0
        self.coulomb = {(0,1,1,0):U}
        #
        vbias, temp, dband = 5.0, 50.0, 1000.0
        self.mulst = [vbias/2, -vbias/2, vbias/2, -vbias/2]
        self.tlst = [temp, temp, temp, temp]
        self.dlst = [dband, dband, dband, dband]
        #
        self.nsingle, self.nleads = 2, 4
        self.kpnt = np.power(2, 9)

class Calcs(object):
    def __init__(self):
        pass

def save_Builder_double_dot_spinful(fname='data_builder.py'):
    p = Parameters_double_dot_spinful()
    #data = {}
    kerns = ['Pauli', 'Redfield', '1vN', 'Lindblad', 'pyPauli', 'pyRedfield', 'py1vN', 'pyLindblad']
    #kerns = ['Pauli']
    itypes = [0, 1, 2]
    data = 'data = {\n'
    dataR = ''
    for kerntype, itype in itertools.product(kerns, itypes):
        if not ( kerntype in {'Pauli', 'pyPauli', 'Lindblad', 'pyLindblad'} and itype in [0, 1] ):
            system = Builder(p.nsingle, p.hsingle, p.coulomb, p.nleads, p.tleads, p.mulst, p.tlst, p.dlst,
                             kerntype=kerntype, itype=itype)
            system.solve()
            attr = kerntype+str(itype)
            data = data+' '*4+'\''+attr+'current\': '+str(system.current.tolist())+',\n'
            data = data+' '*4+'\''+attr+'energy_current\': '+str(system.energy_current.tolist())
            data = data + ('\n    }' if kerntype is 'pyLindblad' and itype is 2 else ',\n' )
    #
    with open(fname, 'w') as f:
        f.write(data)

def test_Builder_double_dot_spinful():
    from qmeq.tests.data_builder import data
    p = Parameters_double_dot_spinful()
    calcs = Calcs()

    # Check if the results agree with previously calculated data
    kerns = ['Pauli', 'Redfield', '1vN', 'Lindblad']
    kerns += ['pyPauli', 'pyRedfield', 'py1vN', 'pyLindblad'] if CHECK_PY else []
    itypes = [0, 1, 2]
    for kerntype, itype in itertools.product(kerns, itypes):
        if not ( kerntype in {'Pauli', 'pyPauli', 'Lindblad', 'pyLindblad'} and itype in [0, 1] ):
            system = Builder(p.nsingle, p.hsingle, p.coulomb, p.nleads, p.tleads, p.mulst, p.tlst, p.dlst,
                             kerntype=kerntype, itype=itype)
            system.solve()
            attr = kerntype+str(itype)
            setattr(calcs, attr, system)
            #
            if PRNTQ:
                print(kerntype, itype)
                print(system.tt.current)
                print( data[attr+'current'] )
                print(system.tt.energy_current)
                print( data[attr+'energy_current'] )
                print( norm(system.current - data[attr+'current']) )
                print( norm(system.energy_current - data[attr+'energy_current']) )
            #
            assert norm(system.current - data[attr+'current']) < EPS
            assert norm(system.energy_current - data[attr+'energy_current']) < EPS

    # Check least-squares solution with non-square matrix, i.e., symq=False
    for kerntype in kerns:
        system = Builder(p.nsingle, p.hsingle, p.coulomb, p.nleads, p.tleads, p.mulst, p.tlst, p.dlst,
                         kerntype=kerntype, itype=2, symq=False)
        system.solve()
        attr = kerntype+str(itype)
        for param in ['current', 'energy_current']:
            assert norm(getattr(system, param) - getattr(getattr(calcs, attr), param)) < EPS

    # Check matrix-free methods
    kerns = ['Pauli', 'Redfield', '1vN', 'Lindblad']
    kerns += ['pyPauli', 'pyRedfield', 'py1vN', 'pyLindblad'] if CHECK_PY else []
    for kerntype in kerns:
        system = Builder(p.nsingle, p.hsingle, p.coulomb, p.nleads, p.tleads, p.mulst, p.tlst, p.dlst,
                         kerntype=kerntype, itype=2, mfreeq=True)
        system.solve()
        attr = kerntype+str(itype)
        for param in ['current', 'energy_current']:
            assert norm(getattr(system, param) - getattr(getattr(calcs, attr), param)) < 1e-4

    # Check results with different indexing
    kerns = ['Pauli', 'Redfield', '1vN', 'Lindblad']
    kerns += ['pyPauli', 'pyRedfield', 'py1vN', 'pyLindblad'] if CHECK_PY else []
    indexings = ['Lin', 'charge', 'sz', 'ssq']
    for kerntype, indexing in itertools.product(kerns, indexings):
        system = Builder(p.nsingle, p.hsingle, p.coulomb, p.nleads, p.tleads, p.mulst, p.tlst, p.dlst,
                         kerntype=kerntype, itype=2, indexing=indexing)
        system.solve()
        attr = kerntype+str(itype)
        for param in ['current', 'energy_current']:
            assert norm(getattr(system, param) - getattr(getattr(calcs, attr), param)) < EPS

def test_Builder_double_dot_spinless_2vN():
    data_current = {'2vN': [0.18472226147540757, -0.1847222614754047]}
    data_energy_current = {'2vN': [0.2828749942707809, -0.28333373130210493]}
    #
    kerns = ['2vN']
    kerns += ['py2vN'] if CHECK_PY else []
    indexings = ['Lin', 'charge']
    p = Parameters_double_dot_spinless()
    for kerntype, indexing in itertools.product(kerns, indexings):
        system = Builder(p.nsingle, p.hsingle, p.coulomb, p.nleads, p.tleads, p.mulst, p.tlst, p.dlst,
                         kerntype=kerntype, indexing=indexing, kpnt=p.kpnt)
        system.solve(niter=5)
        assert norm(system.current - data_current['2vN']) < EPS
        assert norm(system.energy_current - data_energy_current['2vN']) < EPS

def test_Builder_single_orbital_spinful():
    data_current = {'Pauli': [0.08368833245372147, -0.08368833245372037, 0.08368833245372147, -0.08368833245372037],
                    '2vN':   [0.0735967870902393, -0.07359678709023731, 0.07359678709023965, -0.07359678709023706]}
    data_energy_current = {'Pauli': [0.1254772916094682, -0.12547729160946774, 0.1254772916094682, -0.12547729160946774],
                           '2vN':   [0.8745228418442359, -0.8753178905524672, 0.8745228418440387, -0.8753178905524854]}
    #
    kerns = ['Pauli', 'Redfield', '1vN', 'Lindblad']
    kerns += ['pyPauli', 'pyRedfield', 'py1vN', 'pyLindblad'] if CHECK_PY else []
    indexings = ['Lin', 'charge', 'sz', 'ssq']
    itypes = [0, 1, 2]
    p = Parameters_single_orbital_spinful()
    for kerntype, indexing, itype in itertools.product(kerns, indexings, itypes):
        system = Builder(p.nsingle, p.hsingle, p.coulomb, p.nleads, p.tleads, p.mulst, p.tlst, p.dlst,
                         kerntype=kerntype, indexing=indexing, itype=itype)
        system.solve()
        assert norm(system.current - data_current['Pauli']) < EPS
        assert norm(system.energy_current - data_energy_current['Pauli']) < EPS

    kerns = ['2vN']
    kerns += ['py2vN'] if CHECK_PY else []
    indexings = ['Lin', 'charge']
    for kerntype, indexing in itertools.product(kerns, indexings):
        system = Builder(p.nsingle, p.hsingle, p.coulomb, p.nleads, p.tleads, p.mulst, p.tlst, p.dlst,
                         kerntype=kerntype, indexing=indexing, kpnt=p.kpnt)
        system.solve(niter=5)
        assert norm(system.current - data_current['2vN']) < EPS
        assert norm(system.energy_current - data_energy_current['2vN']) < EPS

def test_Builder_coulomb_symmetry_spin():

    def intas(m, l):
        m = m+1
        l = l+1
        return np.exp(1j*(np.sqrt(l)-np.sqrt(m)))*np.sqrt(1/(np.sqrt(m)+np.sqrt(l)))

    def Vmnkl(m, n, k, l):
        return intas(m, l)*intas(n, k)

    nsingle = 4
    # Spinless, m_less_n=False, herm_c=False
    coulomb0 = {}
    for m, n, k, l in itertools.product(range(nsingle), repeat=4):
        coulomb0.update({(m, n, k, l): Vmnkl(m,n,k,l)/2})

    sys0_spinless = Builder(nsingle=nsingle, coulomb=coulomb0, mtype_qd=complex, m_less_n=False, herm_c=False)
    sys0_spinless.solve(masterq=False)

    # Spinless, m_less_n=False, herm_c=True
    coulomb1 = {}
    for m, n, k, l in itertools.product(range(nsingle), repeat=4):
        if not ((l, k, n, m) in coulomb1.keys()):
            coulomb1.update({(m, n, k, l): Vmnkl(m,n,k,l)/2})

    sys1_spinless = Builder(nsingle=nsingle, coulomb=coulomb1, mtype_qd=complex, m_less_n=False, herm_c=True)
    sys1_spinless.solve(masterq=False)

    # Spinless, m_less_n=True, herm_c=False
    coulomb2 = {}
    for m, n, k, l in itertools.product(range(nsingle), repeat=4):
        if m <= n:
            coulomb2.update({(m, n, k, l): Vmnkl(m,n,k,l)})

    sys2_spinless = Builder(nsingle=nsingle, coulomb=coulomb2, mtype_qd=complex, m_less_n=True, herm_c=False)
    sys2_spinless.solve(masterq=False)

    # Spinless, m_less_n=True, herm_c=True
    coulomb3 = {}
    for m, n, k, l in itertools.product(range(nsingle), repeat=4):
        if m <= n and not ((l, k, n, m) in coulomb3.keys()):
            coulomb3.update({(m, n, k, l): Vmnkl(m,n,k,l)})

    sys3_spinless = Builder(nsingle=nsingle, coulomb=coulomb3, mtype_qd=complex, m_less_n=True, herm_c=True)
    sys3_spinless.solve(masterq=False)

    assert sum(abs(sys1_spinless.Ea-sys0_spinless.Ea)) < EPS
    assert sum(abs(sys2_spinless.Ea-sys0_spinless.Ea)) < EPS
    assert sum(abs(sys3_spinless.Ea-sys0_spinless.Ea)) < EPS

    nsingle = 8
    ns = 4
    # Spinful
    coulomb = {}
    for m, n, k, l in itertools.product(range(ns), repeat=4):
        coulomb.update({(m, n, k, l): Vmnkl(m,n,k,l)/2})
        coulomb.update({(m+ns, n, k, l+ns): Vmnkl(m,n,k,l)/2})
        coulomb.update({(m, n+ns, k+ns, l): Vmnkl(m,n,k,l)/2})
        coulomb.update({(m+ns, n+ns, k+ns, l+ns): Vmnkl(m,n,k,l)/2})

    indexing = 'ssq'
    sys_ref_spinful = Builder(nsingle=nsingle, coulomb=coulomb, mtype_qd=complex, m_less_n=False, herm_c=False, indexing=indexing)
    sys_ref_spinful.solve(masterq=False)
    sys0_spinful = Builder(nsingle=nsingle, coulomb=coulomb0, mtype_qd=complex, m_less_n=False, herm_c=False, symmetry='spin', indexing=indexing)
    sys0_spinful.solve(masterq=False)
    sys1_spinful = Builder(nsingle=nsingle, coulomb=coulomb1, mtype_qd=complex, m_less_n=False, herm_c=True, symmetry='spin', indexing=indexing)
    sys1_spinful.solve(masterq=False)
    sys2_spinful = Builder(nsingle=nsingle, coulomb=coulomb2, mtype_qd=complex, m_less_n=True, herm_c=False, symmetry='spin', indexing=indexing)
    sys2_spinful.solve(masterq=False)
    sys3_spinful = Builder(nsingle=nsingle, coulomb=coulomb3, mtype_qd=complex, m_less_n=True, herm_c=True, symmetry='spin', indexing=indexing)
    sys3_spinful.solve(masterq=False)

    assert sum(abs(sys0_spinful.Ea-sys_ref_spinful.Ea)) < EPS
    assert sum(abs(sys1_spinful.Ea-sys_ref_spinful.Ea)) < EPS
    assert sum(abs(sys2_spinful.Ea-sys_ref_spinful.Ea)) < EPS
    assert sum(abs(sys3_spinful.Ea-sys_ref_spinful.Ea)) < EPS

def serial_triple_dot_coulomb_symmetry_spin():
    # Coulomb matrix elements
    # Intradot terms: u, uex
    # Interdot terms: un, udc, usc
    u, uex, un, udc, usc = 10., 2., 3., -0.5, -0.2

    #----------- Spinless -----------
    nsingle = 5
    dotindex = [0, 0, 1, 1, 2]
    # m, n, k, l
    coulomb = []
    for m, n, k, l in itertools.product(range(nsingle), repeat=4):
        if m == n == k == l:
            coulomb.append([m,m,m,m,u/2]) # Direct
        if dotindex[m] == dotindex[k]:
            if m == n and k == l and m != k:
                coulomb.append([m,m,k,k,uex/2]) # Exchange
        if m != n and k != l:
            # Intradot
            if dotindex[m] == dotindex[n]:
                if m == l and n == k: coulomb.append([m,n,n,m,u/2])   # Direct
                if m == k and n == l: coulomb.append([m,n,m,n,uex/2]) # Exchange
            # Interdot
            # Note that the pairs (n,k) and (m,l) are located at different dots
            if (dotindex[m] == dotindex[l] and
                dotindex[n] == dotindex[k] and
                abs(dotindex[m]-dotindex[n]) == 1):
                if m == l and n == k:
                    coulomb.append([m,n,n,m,un/2])   # Direct
                if n == k and m != l:
                    sgn = dotindex[m]-dotindex[n]
                    coulomb.append([m,n,n,l,udc/2*sgn])  # Charge-dipole
                if n != k and m == l:
                    sgn = dotindex[n]-dotindex[m]
                    coulomb.append([m,n,k,m,udc/2*sgn])  # Charge-dipole
                if n != k and m != l:
                    coulomb.append([m,n,k,l,usc/2])  # Charge-quadrupole
    coulomb0 = coulomb

    coulomb1 = {(0,0,0,0):u, (1,1,1,1):u, (2,2,2,2):u, (3,3,3,3):u, (4,4,4,4):u,
                (0,1,1,0):u, (2,3,3,2):u,
                #
                (0,0,1,1):uex, (2,2,3,3):uex,
                (0,1,0,1):uex, (2,3,2,3):uex,
                #
                (0,2,2,0):un, (0,3,3,0):un,
                (1,2,2,1):un, (1,3,3,1):un,
                (2,4,4,2):un, (3,4,4,3):un,
                #
                (0,2,2,1):-udc, (0,3,3,1):-udc, (2,4,4,3):-udc,
                (0,2,3,0):+udc, (1,2,3,1):+udc,
                #
                (0,2,3,1):usc, (0,3,2,1):usc,
                # Conjugated terms
                (1,1,0,0):uex, (3,3,2,2):uex,
                #
                (1,2,2,0):-udc, (1,3,3,0):-udc, (3,4,4,2):-udc,
                (0,3,2,0):+udc, (1,3,2,1):+udc,
                #
                (1,3,2,0):usc, (1,2,3,0):usc
                }

    coulomb2 = {(0,0,0,0):u, (1,1,1,1):u, (2,2,2,2):u, (3,3,3,3):u, (4,4,4,4):u,
                (0,1,1,0):u, (2,3,3,2):u,
                #
                (0,0,1,1):uex, (2,2,3,3):uex,
                (0,1,0,1):uex, (2,3,2,3):uex,
                #
                (0,2,2,0):un, (0,3,3,0):un,
                (1,2,2,1):un, (1,3,3,1):un,
                (2,4,4,2):un, (3,4,4,3):un,
                #
                (0,2,2,1):-udc, (0,3,3,1):-udc, (2,4,4,3):-udc,
                (0,2,3,0):+udc, (1,2,3,1):+udc,
                #
                (0,2,3,1):usc, (0,3,2,1):usc
                }

    sys0_spinless = qmeq.Builder(nsingle=5, coulomb=coulomb0, symmetry='n', m_less_n=False, herm_c=False)
    sys0_spinless.solve(masterq=False)
    sys1_spinless = qmeq.Builder(nsingle=5, coulomb=coulomb1, symmetry='n', m_less_n=True, herm_c=False)
    sys1_spinless.solve(masterq=False)
    sys2_spinless = qmeq.Builder(nsingle=5, coulomb=coulomb2, symmetry='n', m_less_n=True, herm_c=True)
    sys2_spinless.solve(masterq=False)

    assert sum(abs(sys1_spinless.Ea-sys0_spinless.Ea)) < EPS
    assert sum(abs(sys2_spinless.Ea-sys0_spinless.Ea)) < EPS

    #----------- Spinful -----------
    nsingle = 10
    nssl = nsingle//2
    dotindex = [0, 0, 1, 1, 2, 0, 0, 1, 1, 2]
    # m, n, k, l
    coulomb = []
    for m, n, k, l in itertools.product(range(nsingle), repeat=4):
        if m != n and k != l and m//nssl == l//nssl and n//nssl == k//nssl:
            # Intradot
            if dotindex[m] == dotindex[n]:
                if m == l and n == k: coulomb.append([m,n,n,m,u/2]) # Direct
                if m == k and n == l:
                    coulomb.append([m,n,m,n,uex/2]) # Exchange
                    if m+nssl < nsingle:
                        coulomb.append([m,n+nssl,m+nssl,n,uex/2])
                        coulomb.append([m+nssl,n,m,n+nssl,uex/2])
                        coulomb.append([m,m+nssl,n+nssl,n,uex/2])
                        coulomb.append([m+nssl,m,n,n+nssl,uex/2])
            # Interdot
            # Note that the pairs (n,k) and (m,l) are located at different dots
            if (dotindex[m] == dotindex[l] and
                dotindex[n] == dotindex[k] and
                abs(dotindex[m]-dotindex[n]) == 1):
                if m == l and n == k:
                    coulomb.append([m,n,n,m,un/2])   # Direct
                if n == k and m != l:
                    sgn = dotindex[m]-dotindex[n]
                    coulomb.append([m,n,n,l,udc/2*sgn])  # Charge-dipole
                if n != k and m == l:
                    sgn = dotindex[n]-dotindex[m]
                    coulomb.append([m,n,k,m,udc/2*sgn])  # Charge-dipole
                if n != k and m != l:
                    coulomb.append([m,n,k,l,usc/2])  # Charge-quadrupole

    indexing='ssq'
    sys_ref_spinful = qmeq.Builder(nsingle=10, coulomb=coulomb, indexing=indexing)
    sys_ref_spinful.solve(masterq=False)
    sys0_spinful = qmeq.Builder(nsingle=10, coulomb=coulomb0, symmetry='spin', m_less_n=False, indexing=indexing)
    sys0_spinful.solve(masterq=False)
    sys1_spinful = qmeq.Builder(nsingle=10, coulomb=coulomb1, symmetry='spin', m_less_n=True, herm_c=False, indexing=indexing)
    sys1_spinful.solve(masterq=False)
    sys2_spinful = qmeq.Builder(nsingle=10, coulomb=coulomb2, symmetry='spin', m_less_n=True, herm_c=True, indexing=indexing)
    sys2_spinful.solve(masterq=False)

    assert sum(abs(sys0_spinful.Ea-sys_ref_spinful.Ea)) < 10*EPS
    assert sum(abs(sys1_spinful.Ea-sys_ref_spinful.Ea)) < 10*EPS
    assert sum(abs(sys2_spinful.Ea-sys_ref_spinful.Ea)) < 10*EPS

def test_Builder_serial_triple_dot_coulomb_symmetry_spin():
    if SERIAL_TRIPLE_DOT:
        serial_triple_dot_coulomb_symmetry_spin()



import numpy as np
from numpy.linalg import norm
from scipy import exp
from scipy.integrate import quad
from qmeq.specfunc import *
import qmeq

try:
    from qmeq.specfuncc import *
except:
    print("Cannot import Cython compiled module qmeq.specfuncc.")
    c_fermi_func = fermi_func
    c_func_pauli = func_pauli
    c_func_1vN = func_1vN

EPS = 1e-14
EPS2 = 1e-11

def test_fermi_func():
    for f in [fermi_func, c_fermi_func]:
        assert f(0) == 0.5
        assert norm(f(2)+f(-2) - 1.0) < EPS

def test_func_pauli():
    for f in [func_pauli, c_func_pauli]:
        Ecb, mu, T, Dm, Dp, itype = 0, 0, 1, -5, 5, 0
        assert norm(f(Ecb, mu, T, Dm, Dp, itype) - np.pi) < EPS
        fm = f(-2, mu, T, Dm, Dp, itype)
        fp = f(+2, mu, T, Dm, Dp, itype)
        assert norm(fm+fp - 2*np.pi*np.ones(2)) < EPS
        #
        for itype in {0, 2, 4}:
            Ecb = 4.99
            assert norm( f(Ecb, mu, T, Dm, Dp, itype) - [0.04247219960243555, 6.240713107577151] ) < EPS
            Ecb = 5.01
            assert f(Ecb, mu, T, Dm, Dp, itype).tolist() == [0., 0.]
        for itype in {1, 3}:
            assert norm( f(Ecb, mu, T, Dm, Dp, itype) - [0.04163676679420959, 6.241548540385377] ) < EPS

def test_func_1vN():

    def test_rez(Ecb, mu, T, Dm, Dp):
        def f0(x): return 1/(exp((x-mu)/T)+1)
        def f1(x): return -(1-1/(exp((x-mu)/T)+1))
        def f2(x): return x/(exp((x-mu)/T)+1)
        def f3(x): return -x*(1-1/(exp((x-mu)/T)+1))
        cur0, err = quad(f0, Dm, Dp, weight='cauchy', wvar=Ecb, epsabs=1.0e-6, epsrel=1.0e-6, limit=10000)
        cur1, err = quad(f1, Dm, Dp, weight='cauchy', wvar=Ecb, epsabs=1.0e-6, epsrel=1.0e-6, limit=10000)
        en0, err = quad(f2, Dm, Dp, weight='cauchy', wvar=Ecb, epsabs=1.0e-6, epsrel=1.0e-6, limit=10000)
        en1, err = quad(f3, Dm, Dp, weight='cauchy', wvar=Ecb, epsabs=1.0e-6, epsrel=1.0e-6, limit=10000)
        return np.array([cur0, cur1, en0, en1])

    for f in [func_1vN, c_func_1vN]:
        Ecb, mu, T, Dm, Dp = 0.5, 0.3, 1.0, -10., 15.
        rr0 = test_rez(Ecb, mu, T, Dm, Dp)
        rr1 = np.array([-2.4197253517177657, -2.8251904598259303, 8.790137324141117, -16.412595229912966])
        ri = np.array([-1.4142382069390027, 1.7273544466507904, -0.7071191034695014, 0.8636772233253952])
        #
        r0 = f(Ecb, mu, T, Dm, Dp, 0, 10000)
        r1 = f(Ecb, mu, T, Dm, Dp, 1, 10000)
        r2 = f(Ecb, mu, T, Dm, Dp, 2, 10000)
        assert norm(r0 - rr0-1j*ri) < EPS2
        assert norm(r1 - rr1-1j*ri) < EPS2
        assert norm(r2 - 1j*ri) < EPS2
        for itype in {0, 2}:
            r0 = f(Dp+0.1, mu, T, Dm, Dp, itype, 10000)
            assert norm(r0.imag - np.zeros(4)) < EPS
            r0 = f(Dm-0.1, mu, T, Dm, Dp, itype, 10000)
            assert norm(r0.imag - np.zeros(4)) < EPS



"""Module containing Builder class."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy

from .aprclass import Approach
from .indexing import StateIndexingDM
from .indexing import StateIndexingDMc
from .qdot import QuantumDot
from .leadstun import LeadsTunneling

from .various import get_phi0
from .various import get_phi1
from .various import print_state
from .various import print_all_states
from .various import sort_eigenstates
from .various import remove_coherences
from .various import remove_states
from .various import use_all_states

#-----------------------------------------------------------
# Python modules

from .approach.pauli import Approach_pyPauli
from .approach.lindblad import Approach_pyLindblad
from .approach.redfield import Approach_pyRedfield
from .approach.neumann1 import Approach_py1vN
from .approach.neumann2 import Approach_py2vN

# Cython compiled modules

try:
    from .approach.c_pauli import Approach_Pauli
    from .approach.c_lindblad import Approach_Lindblad
    from .approach.c_redfield import Approach_Redfield
    from .approach.c_neumann1 import Approach_1vN
    from .approach.c_neumann2 import Approach_2vN
except:
    print("Cannot import Cython compiled modules for the approaches.")
    Approach_Pauli = Approach_pyPauli
    Approach_Lindblad = Approach_pyLindblad
    Approach_Redfield = Approach_pyRedfield
    Approach_1vN = Approach_py1vN
    Approach_2vN = Approach_py2vN
#-----------------------------------------------------------

class Builder(object):
    """
    Class for building the system for stationary transport calculations.

    Attributes
    ----------
    nsingle : int
        Number of single-particle states.
    hsingle : dict, list, or array
        Dictionary, list, or array corresponding to single-particle hopping (tunneling) Hamiltonian.
        On input list or array gets converted to dictionary.
    coulomb : list, dict, or array
        Dictionary, list or array containing coulomb matrix elements.
        For dictionary:    coulomb[(m, n, k, l)] = U, where m, n, k, l are the state labels.
        For list or array: coulomb[i] is list of the format [m, n, k, l, U].
        U is the strength of the coulomb interaction between the states (m, n, k, l).
        Note that only the matrix elements k>l, n>m have to be specified.
        On input list or array gets converted to dictionary.
    nleads : int
        Number of the leads.
    tleads : dict, list, or array
        Dictionary, list, or numpy array defining single particle tunneling amplitudes.
        numpy array has to be nleads by nsingle.
    mulst : dict, list, or array
        Dictionary, list, or array containing chemical potentials of the leads.
    tlst : dict, list, or array
        Dictionary, list, or array containing temperatures of the leads.
    dband : float, dict, list, or array
        Float, dictionary, list, or array determining bandwidths of the leads.
    indexing : str
        String determining type of the indexing. Possible values are 'Lin', 'charge', 'sz', 'ssq'.
        Note that 'sz' indexing for Fock states is used for 'ssq' indexing, with
        additional specification of eigensates in Fock basis.
    symmetry : str
        String determining that the states will be augmented by the symmetry.
        Possible value is 'spin'.
    kpnt : int
        Number of energy grid points on which 2vN approach equations are solved.
    kerntype : string, Approach class
        String describing what master equation approach to use.
        For Approach class the possible values are 'Pauli', '1vN', 'Redfield', 'Lindblad', \
                                                   'pyPauli', 'py1vN', 'pyRedfield', 'pyLindblad'.
        For Approach2vN class the possible values are '2vN' and 'py2vN'.
        The approaches with 'py' in front are not compiled using cython.
        kerntype can also be an Approach class defining a custom approach.
    symq : bool
        For symq=False keep all equations in the kernel, and the matrix is of size N by N+1.
        For symq=True replace one equation by the normalisation condition,
        and the matrix is square N by N.
    norm_row : int
        If symq=True this row will be replaced by normalisation condition in the kernel matrix.
    solmethod : string
        String specifying the solution method of the equation L(Phi0)=0.
        The possible values are matrix inversion 'solve' and least squares 'lsqr'.
        Method 'solve' works only when symq=True.
        For matrix free methods (used when mfreeq=True) the possible values are
        'krylov', 'broyden', etc.
    itype : int
        Type of integral for first order approach calculations.
        itype=0: the principal parts are evaluated using Fortran integration package QUADPACK \
                 routine dqawc through SciPy.
        itype=1: the principal parts are kept, but approximated by digamma function valid for \
                 large bandwidht D.
        itype=2: the principal parts are neglected.
        itype=3: the principal parts are neglected and infinite bandwidth D is assumed.
    dqawc_limit : int
        For itype=0 dqawc_limit determines the maximum number of subintervals
        in the partition of the given integration interval.
    mfreeq : bool
        If mfreeq=True the matrix free solution method is used for first order methods.
    phi0_init : array
        For mfreeq=True the initial value of zeroth order density matrix elements.
    mtype_qd : float or complex
        Type for the many-body quantum dot Hamiltonian matrix.
    mtype_leads : float or complex
        Type for the many-body tunneling matrix Tba.
    symmetry : str
        String determining if the states will be augmented by a symmetry.
        Possible value is 'spin'.
    herm_hs : bool
        When herm_hs=True the conjugated elements are added automatically to many-body Hamiltonian
        from hsingle. For example, when hsingle={(0,1): val} the element {(1,0): val.conjugate()}
        will also be added.
    herm_c : bool
        When herm_c=True the conjugated elements are added automatically to many-body Hamiltonian
        from coulomb. For example, when coulomb={(0,2,3,1): val} the element
        {(1,3,2,0): val.conjugate()} will also be added. However, we note that for m_less_n=True
        if coulomb={(0,2,1,3): val} the element (3,1,2,0) will not be included.
    m_less_n : bool
        When m_less_n=True the coulomb matrix element (m, n, k, l) has to have m<n.
    qd : QuantumDot
        QuantumDot object.
    leads : LeadsTunneling
        LeadsTunneling object.
    si : StateIndexingDM
        StateIndexingDM object.
    appr : Approach or Approach2vN.
        Approach or Approach2vN object.
    funcp : FunctionProperties
        FunctionProperties object.
    Ea : array
        nmany by 1 array containing many-body Hamiltonian eigenvalues.
    Tba : array
        nleads by nmany by nmany array, which contains many-body tunneling amplitude matrix,
        which is used in calculations.
    kern : array
        Kernel (Liouvillian) representing the master equation.
    phi0 : array
        Values of zeroth order density matrix elements.
    phi1 : array
        Values of first order density matrix elements
        stored in nleads by ndm1 numpy array.
    current : array
        Values of the current having nleads entries.
    energy_current : array
        Values of the energy current having nleads entries.
    heat_current : array
        Values of the heat current having nleads entries.
    niter : int
        Number of iterations performed when solving integral equation for 2vN approach.
    iters : Iterations2vN
        Iterations2vN object, which is present just for 2vN approach.
    """

    def __init__(self, nsingle=0, hsingle={}, coulomb={},
                       nleads=0, tleads={}, mulst={}, tlst={}, dband={},
                       indexing='n', kpnt=None,
                       kerntype='Pauli', symq=True, norm_row=0, solmethod='n',
                       itype=0, dqawc_limit=10000, mfreeq=False, phi0_init=None,
                       mtype_qd=complex, mtype_leads=complex,
                       symmetry='n', herm_hs=True, herm_c=False, m_less_n=True):

        if indexing is 'n':
            if symmetry is 'spin' and kerntype not in {'py2vN', '2vN'}:
                indexing = 'ssq'
            else:
                indexing = 'charge'

        if not indexing in {'Lin', 'charge', 'sz', 'ssq'}:
            print("WARNING: Allowed indexing values are: \'Lin\', \'charge\', \'sz\', \'ssq\'. "+
                  "Using default indexing=\'charge\'.")
            indexing = 'charge'

        if not itype in {0,1,2,3}:
            print("WARNING: itype needs to be 0, 1, 2, or 3. Using default itype=0.")
            itype = 0

        if isinstance(kerntype, str):
            if not kerntype in {'Pauli', 'Lindblad', 'Redfield', '1vN', '2vN',
                                'pyPauli', 'pyLindblad', 'pyRedfield', 'py1vN', 'py2vN'}:
                print("WARNING: Allowed kerntype values are: "+
                      "\'Pauli\', \'Lindblad\', \'Redfield\', \'1vN\', \'2vN\', "+
                      "\'pyPauli\', \'pyLindblad\', \'pyRedfield\', \'py1vN\', \'py2vN\'. "+
                      "Using default kerntype=\'Pauli\'.")
                kerntype = 'Pauli'
            self.Approach = globals()['Approach_'+kerntype]
        else:
            if issubclass(kerntype, Approach):
                self.Approach = kerntype
                kerntype = self.Approach.kerntype

        if not indexing in {'Lin', 'charge'} and kerntype in {'py2vN', '2vN'}:
            print("WARNING: For 2vN approach indexing needs to be \'Lin\' or \'charge\'. "+
                  "Using indexing=\'charge\' as a default.")
            indexing = 'charge'

        # Make copies of initialized parameters.
        hsingle = copy.deepcopy(hsingle)
        coulomb = copy.deepcopy(coulomb)
        tleads = copy.deepcopy(tleads)
        mulst = copy.deepcopy(mulst)
        tlst = copy.deepcopy(tlst)
        dband = copy.deepcopy(dband)
        phi0_init = copy.deepcopy(phi0_init)

        self.funcp = FunctionProperties(symq=symq, norm_row=norm_row, solmethod=solmethod,
                                        itype=itype, dqawc_limit=dqawc_limit,
                                        mfreeq=mfreeq, phi0_init=phi0_init,
                                        mtype_qd=mtype_qd, mtype_leads=mtype_leads,
                                        kpnt=kpnt, dband=dband)

        icn = self.Approach.indexing_class_name
        self.si = globals()[icn](nsingle, indexing, symmetry)
        self.qd = QuantumDot(hsingle, coulomb, self.si, herm_hs, herm_c, m_less_n, mtype_qd)
        self.leads = LeadsTunneling(nleads, tleads, self.si, mulst, tlst, dband, mtype_leads)

        self.appr = self.Approach(self)

    def __getattr__(self, item):
        if item in {'solve', 'current', 'energy_current', 'heat_current', 'phi0', 'phi1',
                    'niter', 'iters', 'kern'}:
            return getattr(self.appr, item)
        elif item in {'kpnt', 'symq', 'norm_row', 'solmethod', 'itype', 'dqawc_limit',
                      'mfreeq', 'phi0_init'}:
            return getattr(self.funcp, item)
        elif item in {'tleads', 'mulst', 'tlst', 'dlst', 'Tba'}:
            return getattr(self.leads, item)
        elif item in {'nsingle', 'nleads'}:
            return getattr(self.si, item)
        elif item in {'hsingle', 'coulomb', 'Ea'}:
            return getattr(self.qd, item)
        else:
            #return self.__getattribute__(item)
            return super(Builder, self).__getattribute__(item)

    def __setattr__(self, item, value):
        if item in {'solve', 'current', 'energy_current', 'heat_current', 'phi0', 'phi1',
                    'niter', 'iters', 'kern'}:
            setattr(self.appr, item, value)
        elif item in {'kpnt', 'symq', 'norm_row', 'solmethod', 'itype', 'dqawc_limit',
                      'mfreeq', 'phi0_init'}:
            setattr(self.funcp, item, value)
        elif item in {'tleads', 'mulst', 'tlst', 'dlst', 'Tba'}:
            setattr(self.leads, item, value)
        elif item in {'nsingle', 'nleads'}:
            setattr(self.si, item, value)
        elif item in {'hsingle', 'coulomb', 'Ea'}:
            setattr(self.qd, item, value)
        else:
            super(Builder, self).__setattr__(item, value)

    def change_si(self):
        si = self.si
        icn = self.Approach.indexing_class_name
        if not isinstance(si, globals()[icn]):
            self.si = globals()[icn](si.nsingle, si.indexing, si.symmetry, si.nleads)
            self.qd.si = self.si
            self.leads.si = self.si

    # kerntype
    def get_kerntype(self):
        return self.appr.kerntype
    def set_kerntype(self, value):
        if isinstance(value, str):
            if self.appr.kerntype != value:
                self.Approach = globals()['Approach_'+value]
                self.change_si()
                self.appr = self.Approach(self)
        else:
            if issubclass(value, Approach):
                self.Approach = value
                self.change_si()
                self.appr = self.Approach(self)
    kerntype = property(get_kerntype, set_kerntype)

    # indexing
    def get_indexing(self):
        return self.si.indexing
    def set_indexing(self, value):
        if self.si.indexing != value:
            print('WARNING: Cannot change indexing from \''+self.si.indexing+'\' to \''+value
                  +'\' Consider contructing a new system using \'indexing='+value+'\'')
    indexing = property(get_indexing, set_indexing)

    # dband
    def get_dband(self):
        return self.funcp.dband
    def set_dband(self, value):
        self.funcp.dband = value
        self.leads.change(dlst=dband)
    dband = property(get_dband, set_dband)

    def add(self, hsingle=None, coulomb=None, tleads=None, mulst=None, tlst=None, dlst=None):
        """
        Adds the values to the specified dictionaries and correspondingly redefines
        relevant many-body properties of the system.

        Parameters
        ----------
        hsingle,coulomb,tleads,mulst,tlst,dlst : dict
            Dictionaries describing what values to add.
            For example, tleads[(lead, state)] = value to add.
        """
        if not (hsingle is None and coulomb is None):
            self.qd.add(hsingle, coulomb)
        if not (tleads is None and mulst is None and tlst is None and dlst is None):
            self.leads.add(tleads, mulst, tlst, dlst)

    def change(self, hsingle=None, coulomb=None, tleads=None, mulst=None, tlst=None, dlst=None):
        """
        Changes the values of the specified dictionaries and correspondingly redefines
        relevant many-body properties of the system.

        Parameters
        ----------
        hsingle,coulomb,tleads,mulst,tlst,dlst : dict
            Dictionaries describing what values to change.
            For example, tleads[(lead, state)] = value to change.
        """
        if not (hsingle is None and coulomb is None):
            self.qd.change(hsingle, coulomb)
        if not (tleads is None and mulst is None and tlst is None and dlst is None):
            self.leads.change(tleads, mulst, tlst, dlst)

    def get_phi0(self, b, bp):
        '''
        Get the reduced density matrix element corresponding to
        many-body states b and bp.
        '''
        return get_phi0(self, b, bp)

    def get_phi1(self, l, c, b):
        '''
        Get the energy integrated current amplitudes corresponding to
        lead l and many-body states c and b.
        '''
        return get_phi1(self, l, c, b)

    def print_state(self, b, eps=0.0, prntq=True, filename=None, separator=''):
        '''
        Prints properties of given many-body eigenstate of the quantum dot Hamiltonain
        '''
        print_state(self, b, eps, prntq, filename, separator)

    def print_all_states(self, filename, eps=0.0, separator='', mode='w'):
        '''
        Prints properties of all many-body eigenstates to a file.
        '''
        print_all_states(self, filename, eps, separator, mode)

    def sort_eigenstates(self, srt='n'):
        '''
        Sort many-body states of the system by given order of properties.
        '''
        sort_eigenstates(self, srt=srt)

    def remove_coherences(self, dE):
        '''
        Remove the coherences with energy difference larger than dE.
        '''
        remove_coherences(self, dE)

    def remove_states(self, dE):
        '''
        Remove the states with energy dE larger than the ground state
        for the transport calculations.
        '''
        remove_states(self, dE)

    def use_all_states(self):
        '''
        Use all states for the transport calculations.
        '''
        use_all_states(self)


class FunctionProperties(object):
    """
    Class containing miscellaneous variables for Approach and Approach2vN classes.

    Attributes
    ----------
    symq : bool
        For symq=False keep all equations in the kernel, and the matrix is of size N by N+1.
        For symq=True replace one equation by the normalisation condition,
        and the matrix is square N by N.
    norm_row : int
        If symq=True this row will be replaced by normalisation condition in the kernel matrix.
    solmethod : string
        String specifying the solution method of the equation L(Phi0)=0.
        The possible values are matrix inversion 'solve' and least squares 'lsqr'.
        Method 'solve' works only when symq=True.
        For matrix free methods (used when mfreeq=True) the possible values are
        'krylov', 'broyden', etc.
    itype : int
        Type of integral for first order approach calculations.
        itype=0: the principal parts are evaluated using Fortran integration package QUADPACK \
                 routine dqawc through SciPy.
        itype=1: the principal parts are kept, but approximated by digamma function valid for \
                 large bandwidht D.
        itype=2: the principal parts are neglected.
        itype=3: the principal parts are neglected and infinite bandwidth D is assumed.
    dqawc_limit : int
        For itype=0 dqawc_limit determines the maximum number of subintervals
        in the partition of the given integration interval.
    mfreeq : bool
        If mfreeq=True the matrix free solution method is used for first order methods.
    phi0_init : array
        For mfreeq=True the initial value of zeroth order density matrix elements.
    mtype_qd : float or complex
        Type for the many-body quantum dot Hamiltonian matrix.
    mtype_leads : float or complex
        Type for the many-body tunneling matrix Tba.
    kpnt_left, kpnt_right : int
        Number of points Ek_grid is extended to the left and the right for '2vN' approach.
    ht_ker : array
        Kernel used when performing Hilbert transform using FFT.
        It is generated using specfunc.kernel_fredriksen(n).
    emin, emax : float
        Minimal and maximal energy in the updated Ek_grid generated by neumann2py.get_grid_ext(sys).
        Note that emin<=Dmin and emax>=Dmax.
    dmin, dmax : float
        Bandedge Dmin and Dmax values of the lead electrons.
    ext_fct : float
        Multiplication factor used in neumann2py.get_grid_ext(sys), when determining emin and emax.
    suppress_err : bool
        Determines whether to print the warning when the inversion of the kernel failed.
    """

    def __init__(self, kerntype='2vN', symq=True, norm_row=0, solmethod='n',
                       itype=0, dqawc_limit=10000, mfreeq=False, phi0_init=None,
                       mtype_qd=float, mtype_leads=complex, kpnt=None, dband=None):
        self.kerntype = kerntype
        self.symq = symq
        self.norm_row = norm_row
        self.solmethod = solmethod
        #
        self.itype = itype
        self.dqawc_limit = dqawc_limit
        #
        self.mfreeq = mfreeq
        self.phi0_init = phi0_init
        #
        self.mtype_qd = mtype_qd
        self.mtype_leads = mtype_leads
        #
        self.kpnt = kpnt
        self.dband = dband
        #
        self.kpnt_left = 0
        self.kpnt_right = 0
        self.ht_ker = None
        #
        self.dmin, self.dmax = 0, 0
        self.emin, self.emax = 0, 0
        self.ext_fct = 1.1
        #
        self.suppress_err = False
        self.suppress_wrn = [False]

    def print_error(self, exept):
        if not self.suppress_err:
            print(str(exept))
            print("WARNING: Could not invert the kernel. "+
                  "All the transport channels may be outside the bandwidth. "+
                  "This warning will not be shown again.")
            self.suppress_err = True

    def print_warning(self, i, message):
        if not self.suppress_wrn[i]:
            print(message)
            self.suppress_wrn[i] = True



from qmeq.aprclass import *
import qmeq

def test_Approach2vN_kpnt():
    system = qmeq.Builder(nleads=1, dband={0: 1000}, kpnt=5, kerntype='2vN')
    appr = Approach2vN(system)
    appr.make_Ek_grid()
    assert appr.Ek_grid.tolist() == [-1000, -500, 0, 500, 1000]
    appr.funcp.kpnt = 6
    appr.make_Ek_grid()
    assert appr.Ek_grid.tolist() == [-1000, -600,  -200, 200, 600, 1000]
    #
    system = qmeq.Builder(1, {}, {}, 1, {}, {}, {}, {0: 1000}, kpnt=5, kerntype='2vN')
    system.appr.make_Ek_grid()
    assert system.appr.Ek_grid.tolist() == [-1000, -500, 0, 500, 1000]
    system.kpnt = 6
    system.appr.make_Ek_grid()
    assert system.appr.Ek_grid.tolist() == [-1000, -600,  -200, 200, 600, 1000]

def test_Approach2vN_make_Ek_grid():
    system = qmeq.Builder(nleads=2, dband={0: [-1000, 1000], 1: [-1000, 1000]}, kpnt=5, kerntype='2vN')
    appr = Approach2vN(system)
    appr.make_Ek_grid()
    assert appr.Ek_grid.tolist() == [-1000, -500, 0, 500, 1000]
    appr.leads.change(dlst={0: [-1400, 1000], 1: [-1000, 1000]})
    appr.make_Ek_grid()
    assert appr.Ek_grid.tolist() == [-1400.0, -800.0, -200.0, 400.0, 1000.0]



"""Module containing cython functions, which generate first order Pauli kernel.
   For docstrings see documentation of module pauli."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import itertools

from ..mytypes import doublenp
from ..mytypes import complexnp

from ..specfuncc cimport func_pauli
from ..aprclass import Approach

cimport numpy as np
cimport cython

ctypedef np.uint8_t bool_t
ctypedef np.int_t int_t
ctypedef np.int64_t long_t
ctypedef np.float64_t double_t
ctypedef np.complex128_t complex_t

@cython.boundscheck(False)
def c_generate_paulifct(sys):
    cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
    cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
    si = sys.si
    cdef np.ndarray[double_t, ndim=1] mulst = sys.leads.mulst
    cdef np.ndarray[double_t, ndim=1] tlst = sys.leads.tlst
    cdef np.ndarray[double_t, ndim=2] dlst = sys.leads.dlst
    #
    cdef long_t c, b, cb
    cdef int_t bcharge, ccharge, charge, l
    cdef double_t xcb, Ecb
    cdef int_t nleads = si.nleads
    cdef int_t itype = sys.funcp.itype
    #
    cdef np.ndarray[double_t, ndim=3] paulifct = np.zeros((nleads, si.ndm1, 2), dtype=doublenp)
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    #
    cdef np.ndarray[double_t, ndim=1] rez = np.zeros(2, dtype=doublenp)
    #
    for charge in range(si.ncharge-1):
        ccharge = charge+1
        bcharge = charge
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
            Ecb = E[c]-E[b]
            for l in range(nleads):
                xcb = (Tba[l, b, c]*Tba[l, c, b]).real
                func_pauli(Ecb, mulst[l], tlst[l], dlst[l,0], dlst[l,1], itype, rez)
                paulifct[l, cb, 0] = xcb*rez[0]
                paulifct[l, cb, 1] = xcb*rez[1]
    sys.paulifct = paulifct
    return 0

#---------------------------------------------------------------------------------------------------
# Pauli master equation
#---------------------------------------------------------------------------------------------------
@cython.boundscheck(False)
def c_generate_kern_pauli(sys):
    cdef np.ndarray[double_t, ndim=3] paulifct = sys.paulifct
    si = sys.si
    cdef bint symq = sys.funcp.symq
    cdef long_t norm_rowp = sys.funcp.norm_row
    #
    cdef bool_t bb_bool
    cdef long_t b, bb, a, aa, c, cc, ba, cb
    cdef int_t acharge, bcharge, ccharge, charge, l
    cdef int_t norm_row, last_row
    cdef int_t nleads = si.nleads
    #
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
    cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
    #
    norm_row = norm_rowp if symq else si.npauli
    last_row = si.npauli-1 if symq else si.npauli
    #
    cdef np.ndarray[double_t, ndim=2] kern = np.zeros((last_row+1, si.npauli), dtype=doublenp)
    cdef np.ndarray[double_t, ndim=1] bvec = np.zeros(last_row+1, dtype=doublenp)
    bvec[norm_row] = 1
    for charge in range(si.ncharge):
        acharge = charge-1
        bcharge = charge
        ccharge = charge+1
        for b in si.statesdm[charge]:
            bb = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[b] + shiftlst0[bcharge]]
            bb_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[b] + shiftlst0[bcharge]]
            kern[norm_row, bb] = kern[norm_row, bb] + 1
            if not (symq and bb == norm_row) and bb_bool:
                for a in si.statesdm[charge-1]:
                    aa = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[a] + shiftlst0[acharge]]
                    ba = lenlst[acharge]*dictdm[b] + dictdm[a] + shiftlst1[acharge]
                    for l in range(nleads):
                        kern[bb, bb] = kern[bb, bb] - paulifct[l, ba, 1]
                        kern[bb, aa] = kern[bb, aa] + paulifct[l, ba, 0]
                for c in si.statesdm[charge+1]:
                    cc = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[c] + shiftlst0[ccharge]]
                    cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
                    for l in range(nleads):
                        kern[bb, bb] = kern[bb, bb] - paulifct[l, cb, 0]
                        kern[bb, cc] = kern[bb, cc] + paulifct[l, cb, 1]
    sys.kern = kern
    sys.bvec = bvec
    return 0

@cython.boundscheck(False)
def c_generate_current_pauli(sys):
    cdef np.ndarray[double_t, ndim=1] phi0 = sys.phi0
    cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
    cdef np.ndarray[double_t, ndim=3] paulifct = sys.paulifct
    si = sys.si
    #
    cdef long_t c, cc, b, bb, cb
    cdef int_t bcharge, ccharge, charge, l, nleads
    cdef double_t fct1, fct2
    nleads = si.nleads
    #
    cdef np.ndarray[double_t, ndim=1] current = np.zeros(nleads, dtype=doublenp)
    cdef np.ndarray[double_t, ndim=1] energy_current = np.zeros(nleads, dtype=doublenp)
    #
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
    #
    for charge in range(si.ncharge-1):
        ccharge = charge+1
        bcharge = charge
        for c in si.statesdm[ccharge]:
            cc = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[c] + shiftlst0[ccharge]]
            for b in si.statesdm[bcharge]:
                bb = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[b] + shiftlst0[bcharge]]
                cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
                for l in range(nleads):
                    fct1 = +phi0[bb]*paulifct[l, cb, 0]
                    fct2 = -phi0[cc]*paulifct[l, cb, 1]
                    current[l] = current[l] + fct1 + fct2
                    energy_current[l] = energy_current[l] - (E[b]-E[c])*(fct1 + fct2)
    sys.current = current
    sys.energy_current = energy_current
    sys.heat_current = energy_current - current*sys.leads.mulst
    return 0

@cython.boundscheck(False)
def c_generate_vec_pauli(np.ndarray[double_t, ndim=1] phi0, sys):
    cdef np.ndarray[double_t, ndim=3] paulifct = sys.paulifct
    si = sys.si
    cdef long_t norm_row = sys.funcp.norm_row
    #
    cdef bool_t bb_bool
    cdef long_t b, bb, a, aa, c, cc, ba, cb
    cdef int_t acharge, bcharge, ccharge, charge, l
    cdef int_t nleads = si.nleads
    cdef double_t norm = 0
    #
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
    cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
    #
    cdef np.ndarray[double_t, ndim=1] dphi0_dt = np.zeros(si.npauli, dtype=doublenp)
    for charge in range(si.ncharge):
        acharge = charge-1
        bcharge = charge
        ccharge = charge+1
        for b in si.statesdm[charge]:
            bb = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[b] + shiftlst0[bcharge]]
            bb_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[b] + shiftlst0[bcharge]]
            norm = norm + phi0[bb]
            if bb_bool:
                for a in si.statesdm[charge-1]:
                    aa = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[a] + shiftlst0[acharge]]
                    ba = lenlst[acharge]*dictdm[b] + dictdm[a] + shiftlst1[acharge]
                    for l in range(nleads):
                        dphi0_dt[bb] = dphi0_dt[bb] - paulifct[l, ba, 1]*phi0[bb]
                        dphi0_dt[bb] = dphi0_dt[bb] + paulifct[l, ba, 0]*phi0[aa]
                for c in si.statesdm[charge+1]:
                    cc = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[c] + shiftlst0[ccharge]]
                    cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
                    for l in range(nleads):
                        dphi0_dt[bb] = dphi0_dt[bb] - paulifct[l, cb, 0]*phi0[bb]
                        dphi0_dt[bb] = dphi0_dt[bb] + paulifct[l, cb, 1]*phi0[cc]
    dphi0_dt[norm_row] = norm-1
    return dphi0_dt

class Approach_Pauli(Approach):

    kerntype = 'Pauli'
    generate_fct = c_generate_paulifct
    generate_kern = c_generate_kern_pauli
    generate_current = c_generate_current_pauli
    generate_vec = c_generate_vec_pauli
#---------------------------------------------------------------------------------------------------



from __future__ import print_function

import os
import sys
import numpy as np

try:
    from setuptools import setup, Extension
    #print('installing with setuptools')
except:
    from distutils.core import setup, Extension
    #print('installing with distutils')


def get_ext_modules():
    '''
    Generate C extensions

    1. By default already Cython generated *.c files are used.
    2. If *.c files are not present Cython generates them.
    3. If the option '--cython' is specified Cython generates new *.c files.
    '''

    # Check if *.c files are already there
    file_list = ['qmeq/approach/c_pauli.c',
                 'qmeq/approach/c_lindblad.c',
                 'qmeq/approach/c_redfield.c',
                 'qmeq/approach/c_neumann1.c',
                 'qmeq/approach/c_neumann2.c',
                 'qmeq/specfuncc.c']
    c_files_exist = all([os.path.isfile(f) for f in file_list])

    # Check if --cython option is specified
    if '--cython' in sys.argv:
        use_cython = True
        sys.argv.remove('--cython')
    else:
        use_cython = False

    if c_files_exist and not use_cython:
        cythonize = None
        file_ext = '.c'
        #print('using already Cython generated C files')
    else:
        from Cython.Build import cythonize
        file_ext = '.pyx'
        #print('using cythonize to generate C files')

    ext = [# Pauli
           Extension('qmeq.approach.c_pauli',
                     ['qmeq/approach/c_pauli'+file_ext]),
           # Lindblad
           Extension('qmeq.approach.c_lindblad',
                     ['qmeq/approach/c_lindblad'+file_ext]),
           # Redfield
           Extension('qmeq.approach.c_redfield',
                     ['qmeq/approach/c_redfield'+file_ext]),
           # 1vN
           Extension('qmeq.approach.c_neumann1',
                     ['qmeq/approach/c_neumann1'+file_ext]),
           # 2vN
           Extension('qmeq.approach.c_neumann2',
                     ['qmeq/approach/c_neumann2'+file_ext]),
           # Special functions
           Extension('qmeq.specfuncc',
                     ['qmeq/specfuncc'+file_ext])]

    cext = ext if cythonize is None else cythonize(ext)
    return cext

long_description = open('README.rst').read()

classifiers = ['Development Status :: 5 - Production/Stable',
               'Intended Audience :: Science/Research',
               'License :: OSI Approved :: BSD License',
               'Operating System :: MacOS',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Programming Language :: Cython',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3',
               'Topic :: Scientific/Engineering :: Physics']

setup(name='qmeq',
      version='1.0',
      description=('Package for transport calculations in quantum dots '
                  +'using approximate quantum master equations'),
      long_description=long_description,
      url='http://github.com/gedaskir/qmeq',
      author='Gediminas Kirsanskas',
      author_email='qmeq.package@gmail.com',
      license='BSD 2-Clause',
      classifiers=classifiers,
      packages=['qmeq', 'qmeq/approach', 'qmeq/tests'],
      package_data={'qmeq': ['*.pyx', '*.c', '*.pyd', '*.o', '*.so'],
                    'qmeq/approach': ['*.pyx', '*.c', '*.pyd', '*.o', '*.so']},
      zip_safe=False,
      install_requires=['numpy', 'scipy'],
      include_dirs=[np.get_include()],
      ext_modules=get_ext_modules())
